import tempfile
import subprocess
import json
import os
import pystac
import ubelt as ub
from os.path import join


def egress_item(stac_item, outbucket, aws_base_command):
    if isinstance(stac_item, dict):
        stac_item_dict = stac_item
    elif isinstance(stac_item, pystac.Item):
        stac_item_dict = stac_item.to_dict()
    else:
        raise TypeError("Expecting 'stac_item' to be either a dictionary "
                        "or pystac.Item")

    stac_item_outpath = join(
        outbucket, "{}.json".format(stac_item.id))

    assets_outdir = join(outbucket, stac_item.id)

    for asset_name, asset in stac_item_dict.get('assets', {}).items():
        asset_basename = os.path.basename(asset['href'])

        asset_outpath = join(assets_outdir, asset_basename)

        command = [*aws_base_command, asset['href'], asset_outpath]

        print("Running: {}".format(' '.join(command)))
        # TODO: Manually check return code / output
        subprocess.run(command, check=True)

        # Update feature asset href to point to local outpath
        asset['href'] = asset_outpath

    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            print(json.dumps(stac_item_dict, indent=2), file=f)

        command = [*aws_base_command,
                   temporary_file.name, stac_item_outpath]

        subprocess.run(command, check=True)

    output_stac_item = pystac.Item.from_dict(stac_item_dict)
    output_stac_item.set_self_href(stac_item_outpath)
    return output_stac_item


def ingress_item(feature,
                 outdir,
                 aws_base_command,
                 dryrun,
                 relative=False,
                 virtual=False):
    """
    Originally from the baseline_framework_ingress code; could probably be
    cleaned up.

    FIXME: Something is this is not concurrent-safe
    """
    import subprocess
    from urllib.parse import urlparse
    import pystac
    SENTINEL_PLATFORMS = {'sentinel-2b', 'sentinel-2a'}
    # Adding a reference back to the original STAC
    # item if not already present
    self_link = None
    has_original = False
    for link in feature.get('links', ()):
        if link['rel'] == 'self':
            self_link = link
        elif link['rel'] == 'original':
            has_original = True

    if not has_original and self_link is not None:
        feature.setdefault('links', []).append(
            {'rel': 'original',
             'href': self_link['href'],
             'type': 'application/json'})

    # Should only be added the first time the item is ingressed
    if 'watch:original_item_id' not in feature['properties']:
        feature['properties']['watch:original_item_id'] = feature['id']

    assets = feature.get('assets', {})

    # HTML index page for certain Landsat items, not needed here
    # so remove from assets dict
    if 'index' in assets:
        del assets['index']

    new_assets = {}
    assets_to_remove = set()
    for asset_name, asset in assets.items():
        asset_basename = os.path.basename(asset['href'])

        feature_output_dir = os.path.join(outdir, feature['id'])

        asset_outpath = os.path.join(feature_output_dir, asset_basename)

        asset_href = asset['href']

        try:
            if ('productmetadata' not in assets
               and feature['properties']['platform'] in SENTINEL_PLATFORMS
               and asset_name == 'metadata'):
                asset_outpath = os.path.join(
                    feature_output_dir, 'MTD_TL.xml')

                new_asset = download_mtd_msil1c(
                    feature['properties']['sentinel:product_id'],
                    asset_href, feature_output_dir, aws_base_command,
                    dryrun)

                if new_asset is not None:
                    new_assets['productmetadata'] = new_asset
        except KeyError:
            pass

        local_asset_href = os.path.abspath(asset_outpath)
        if relative:
            local_asset_href = os.path.relpath(asset_outpath, outdir)

        if not dryrun:
            os.makedirs(feature_output_dir, exist_ok=True)

        if os.path.isfile(asset_outpath):
            print('Asset already exists at outpath {!r}, '
                  'not redownloading'.format(asset_outpath))
            # Update feature asset href to point to local outpath
            asset['href'] = local_asset_href
        else:
            # Prefer to pull asset from S3 if available
            parsed_asset_href = urlparse(asset_href)
            if (parsed_asset_href.scheme != 's3'
               and 'alternate' in asset and 's3' in asset['alternate']):
                asset_href_for_download = asset['alternate']['s3']['href']
            else:
                asset_href_for_download = asset_href

            # Need to reparse the href if it switched from http to s3
            parsed_asset_href = urlparse(asset_href_for_download)

            if virtual:
                if parsed_asset_href.scheme == 's3':
                    virtual_href = '/vsis3/{}{}'.format(
                        parsed_asset_href.netloc,
                        parsed_asset_href.path)
                    # print(f'virtual_href={virtual_href}')
                    asset['href'] = virtual_href
                elif parsed_asset_href.scheme in {'http', 'https'}:
                    virtual_href = '/vsicurl/{}://{}{}'.format(
                        parsed_asset_href.scheme,
                        parsed_asset_href.netloc,
                        parsed_asset_href.path)
                    # print(f'virtual_href={virtual_href}')
                    asset['href'] = virtual_href
                else:
                    print("* Unsupported URI scheme '{}' for virtual ingress; "
                          "not updating href: {}".format(
                              parsed_asset_href.scheme, asset_href))
            else:
                try:
                    success = download_file(asset_href_for_download,
                                            asset_outpath,
                                            aws_base_command,
                                            dryrun)
                except subprocess.CalledProcessError:
                    print("* Error * Couldn't download asset from href: '{}', "
                          "removing asset from item!".format(
                              asset_href_for_download))
                    assets_to_remove.add(asset_name)
                    continue
                else:
                    if success:
                        asset['href'] = local_asset_href
                    else:
                        print('Warning unrecognized scheme for asset href: '
                              '{!r}, skipping!'.format(
                                  asset_href_for_download))
                        continue

    for asset_name in assets_to_remove:
        del assets[asset_name]

    for new_asset_name, new_asset in new_assets.items():
        assets[new_asset_name] = new_asset

    item = pystac.Item.from_dict(feature)
    item.set_collection(None)  # Clear the collection if present

    item_href = os.path.join(outdir, feature['id'], feature['id'] + '.json')
    # Transform to absolute
    item_href = os.path.abspath(item_href)
    if relative:
        # Transform to relative if requested
        item_href = os.path.relpath(item_href, outdir)

    item.set_self_href(item_href)
    # print('item = {}'.format(ub.urepr(item.to_dict(), nl=2)))
    return item


def download_mtd_msil1c(product_id,
                        metadata_href,
                        outdir,
                        aws_base_command,
                        dryrun):
    from datetime import datetime as datetime_cls
    import subprocess
    from urllib.parse import urlparse
    # The metadata of the product, which tile is part of, are available in
    # parallel folder (productInfo.json contains the name of the product).
    # This can be found in products/[year]/[month]/[day]/[product name].
    # (https://roda.sentinel-hub.com/sentinel-s2-l1c/readme.html)
    try:
        dt = datetime_cls.strptime(product_id.split('_')[2], '%Y%m%dT%H%M%S')
    except ValueError:
        # Support for older format product ID format, e.g.:
        # "S2A_OPER_PRD_MSIL1C_PDMC_20160413T135705_R065_V20160412T102058_20160412T102058"
        dt = datetime_cls.strptime(product_id.split('_')[7][1:], '%Y%m%dT%H%M%S')

    scheme, netloc, path, *_ = urlparse(metadata_href)
    index = path.find('tiles')
    path = (path[:index] +
            f'products/{dt.year}/{dt.month}/{dt.day}/{product_id}/metadata.xml')
    mtd_msil1c_href = f'{scheme}://{netloc}{path}'
    mtd_msil1c_outpath = os.path.join(outdir, 'MTD_MSIL1C.xml')

    try:
        success = download_file(
            mtd_msil1c_href, mtd_msil1c_outpath, aws_base_command, dryrun)
    except subprocess.CalledProcessError:
        print('* Warning * Failed to download MTD_MSIL1C.xml')
        return None

    if success:
        return {
            'href': mtd_msil1c_outpath,
            'type': 'application/xml',
            'title': 'Product XML metadata',
            'roles': ['metadata']
        }
    else:
        print('Warning unrecognized scheme for asset href: {!r}, '
              'skipping!'.format(mtd_msil1c_href))
        return None


def download_file(href, outpath, aws_base_command, dryrun):
    import subprocess
    from urllib.parse import urlparse
    # TODO: better handling of possible download failure?
    scheme, *_ = urlparse(href)
    verbose = 0
    if scheme == 's3':
        command = [*aws_base_command, href, outpath]
        if verbose > 1:
            print('Running: {}'.format(' '.join(command)))
        # TODO: Manually check return code / output
        subprocess.run(command, check=True)
    elif scheme in {'https', 'http'}:
        if verbose > 1:
            print('Downloading: {!r} to {!r}'.format(href, outpath))
        if not dryrun:
            download_http_file(href, outpath)
    else:
        return False
    return True


def download_http_file(url, outpath):
    import requests
    response = requests.get(url)
    with open(outpath, 'wb') as outf:
        for chunk in response.iter_content(chunk_size=128):
            outf.write(chunk)


class CacheItemOutputS3Wrapper:
    def __init__(self, item_map, outbucket, aws_profile=None):
        self.item_map = item_map
        self.outbucket = outbucket

        aws_cp = AWS_S3_Command('cp')
        aws_cp.update(
            profile=aws_profile,
            no_progress=True,
        )
        self.aws_base_command = aws_cp.finalize()

    def __call__(self, stac_item, *args, **kwargs):
        with tempfile.TemporaryDirectory() as tmpdirname:
            status_file_basename = '{}.done'.format(stac_item['id'])
            status_item_s3_path = os.path.join(
                self.outbucket, 'status', status_file_basename)
            status_item_local_path = os.path.join(
                tmpdirname, status_file_basename)

            try:
                subprocess.run([*self.aws_base_command,
                                status_item_s3_path,
                                status_item_local_path],
                               check=True)
            except subprocess.CalledProcessError:
                pass
            else:
                print("* Item: {} previously processed, not "
                      "re-processing".format(stac_item['id']))
                with open(status_item_local_path) as f:
                    return [json.loads(line) for line in f]

            output_stac_items = self.item_map(stac_item, *args, **kwargs)

            output_status_file = os.path.join(
                tmpdirname, '{}.output.done'.format(stac_item['id']))
            with open(output_status_file, 'w') as outf:
                if isinstance(output_stac_items, dict):
                    print(json.dumps(output_stac_items), file=outf)
                elif isinstance(output_stac_items, pystac.Item):
                    print(json.dumps(output_stac_items.to_dict()), file=outf)
                else:
                    for output_item in output_stac_items:
                        if isinstance(output_item, pystac.Item):
                            print(json.dumps(output_item.to_dict()), file=outf)
                        else:
                            print(json.dumps(output_item), file=outf)

            subprocess.run([*self.aws_base_command,
                            output_status_file,
                            status_item_s3_path], check=True)

            return output_stac_items


def _default_item_selector(stac_item):
    return True


def _default_asset_selector(asset_name, asset):
    return True


class IngressProcessEgressWrapper:
    def __init__(self,
                 item_map,
                 outbucket,
                 aws_base_command,
                 dryrun=False,
                 stac_item_selector=_default_item_selector,
                 asset_selector=_default_asset_selector,
                 skip_egress=False):
        self.item_map = item_map
        self.outbucket = outbucket
        self.aws_base_command = aws_base_command
        self.dryrun = dryrun
        self.stac_item_selector = stac_item_selector
        self.asset_selector = asset_selector
        self.skip_egress = skip_egress

    def __call__(self, stac_item, *args, **kwargs):
        # Assumes that the 'self.item_map' function accepts
        # 'stac_item' and 'working_dir' arguments. TODO: actually
        # check this via introspection
        print("* Processing item: {}".format(stac_item['id']))

        if not self.stac_item_selector(stac_item):
            print("**  STAC item {} did not satisfy selector, not "
                  "processing".format(stac_item['id']))
            return [stac_item]

        with tempfile.TemporaryDirectory() as tmpdirname:
            ingressed_item = ingress_item(
                stac_item,
                os.path.join(tmpdirname, 'ingress'),
                self.aws_base_command,
                self.dryrun,
                relative=False,
                asset_selector=self.asset_selector)

            # Stripping the 'root' link here as it usually refers to
            # the catalog which isn't ingressed when we call
            # ingress_item directly (can throw exceptions when trying
            # to convert to dict or serialize when the catalog is
            # missing)
            ingressed_item.remove_links('root')

            processed_item = self.item_map(
                ingressed_item,
                tmpdirname,
                *args, **kwargs)

            processed_items = []
            if isinstance(processed_item, dict):
                processed_items.append(pystac.Item.from_dict(processed_item))
            elif isinstance(processed_item, pystac.Item):
                processed_items.append(processed_item)
            else:
                # Assume already an iterable of pystac.Item
                processed_items = processed_item

            if self.skip_egress:
                return processed_items

            output_items = []
            for item in processed_items:
                output_items.append(egress_item(item,
                                                self.outbucket,
                                                self.aws_base_command))

            # Returning a list here
            return output_items


def download_region(input_region_path,
                    output_region_path,
                    aws_profile=None,
                    strip_nonregions=False,
                    ensure_comments=False):
    from geowatch.utils.util_fsspec import FSPath
    from geowatch.geoannots import geomodels

    # TODO: handle aws_profile
    assert aws_profile is None, 'unhandled'

    input_region_path = FSPath.coerce(input_region_path)
    with input_region_path.open('r') as file:
        out_region_data = json.load(file)

    region = geomodels.RegionModel.coerce(out_region_data)
    region.fixup()

    if strip_nonregions:
        region.strip_body_features()

    if ensure_comments:
        region.ensure_comments()

    with open(output_region_path, 'w') as f:
        print(region.dumps(indent=2), file=f)

    return output_region_path


def determine_region_id(region_fpath):
    """
    Args:
        region_fpath (str | PathLike):
            the path to a region model geojson file

    Returns:
        str | None : the region id if we can find one
    """
    region_id = None
    with open(region_fpath, 'r') as file:
        region_data = json.load(file)
        for feature in region_data.get('features', []):
            props = feature['properties']
            if props['type'] == 'region':
                region_id = props.get('region_id', props.get('region_model_id'))
                break
    return region_id


class AWS_S3_Command:
    """
    Helper to build and execute AWS S3 bash commands

    Note:
        probably should use fsspec instead of this in most cases.

    References:
        https://docs.aws.amazon.com/cli/latest/reference/s3/

    Example:
        >>> from geowatch.utils.util_framework import *  # NOQA
        >>> self = AWS_S3_Command('ls', 's3://foo/bar')
        >>> self.update(profile='myprofile')
        >>> print(self.finalize())
        ['aws', 's3', '--profile', 'myprofile', 'ls', 's3://foo/bar']
        >>> self = AWS_S3_Command('cp', 's3://foo/bar', '/foo/bar', quiet=True, no_progress=True, color='auto')
        >>> print(self.finalize())
        ['aws', 's3', '--quiet', '--no-progress', '--color', 'auto', 'cp', 's3://foo/bar', '/foo/bar']

    Example:
        >>> # Reuse the same command object with different positional args
        >>> aws_cmd = AWS_S3_Command('cp')
        >>> aws_cmd.update(
        >>>     profile='myprof',
        >>>     only_show_errors=True
        >>> )
        >>> aws_cmd.args = ['s3://data1', '/local/data1']
        >>> print(aws_cmd.finalize())
        ['aws', 's3', '--only-show-errors', '--profile', 'myprof', 'cp', 's3://data1', '/local/data1']
        >>> # Set the `args` attribute to get a new command while keeping
        >>> # existing options.
        >>> aws_cmd.update(recursive=True)
        >>> aws_cmd.args = ['s3://data2', '/local/data2']
        >>> print(aws_cmd.finalize())
        ['aws', 's3', '--only-show-errors', '--recursive', '--profile', 'myprof', 'cp', 's3://data2', '/local/data2']

    Example:
        >>> # There is no need to specify the entire command. If you want
        >>> # to simply build a command prefix, then that works too.
        >>> aws_cmd = AWS_S3_Command('cp', profile='myprof', aws_storage_class='foobar')
        >>> print(aws_cmd.finalize())
        ['aws', 's3', '--profile', 'myprof', '--aws-storage-class', 'foobar', 'cp']
    """

    # Register known options for known commands
    # TODO: multi values

    cmd_known_flags = {}
    cmd_known_keyvals = {}

    cmd_known_flags['ls'] = [
        'recursive',
        'human-readable',
        'summarize',
        'debug',
        'no-verify-ssl',
        'no-paginate',
        'no-sign-request',
    ]

    cmd_known_keyvals['ls'] = [
        'endpoint-url',
        'page-size',
        'request-payer',
        'output',
        'query',
        'profile',
        'region',
        'version',
        'color',
        'ca-bundle',
        'cli-read-timeout',
        'cli-connect-timeout',
    ]

    cmd_known_flags['sync'] = [
        'dryrun',
        'quiet',
        'follow-symlinks',
        'no-follow-symlinks',
        'no-guess-mime-type',
        'only-show-errors',
        'no-progress',
        'ignore-glacier-warnings',
        'force-glacier-transfer',
        'size-only',
        'exact-timestamps',
        'delete',
        'debug',
        'no-verify-ssl',
        'no-paginate',
        'no-sign-request',
    ]

    cmd_known_keyvals['sync'] = [
        'include',
        'exclude',
        'acl',
        'sse',
        'sse-c',
        'sse-c-key',
        'sse-kms-key-id',
        'sse-c-copy-source',
        'sse-c-copy-source-key',
        'storage-class',
        'grants',
        'website-redirect',
        'content-type',
        'cache-control',
        'content-disposition',
        'content-encoding',
        'content-language',
        'expires',
        'source-region',
        'page-size',
        'request-payer',
        'metadata',
        'metadata-directive',
        'endpoint-url',
        'output',
        'query',
        'profile',
        'region',
        'version',
        'color',
        'ca-bundle',
        'cli-read-timeout',
        'cli-connect-timeout',
    ]

    cmd_known_flags['cp'] = [
        'dryrun',
        'quiet',
        'follow-symlinks',
        'no-follow-symlinks',
        'no-guess-mime-type',
        'only-show-errors',
        'no-progress',
        'ignore-glacier-warnings',
        'force-glacier-transfer',
        'recursive',
        'debug',
        'no-verify-ssl',
        'no-paginate',
    ]

    cmd_known_keyvals['cp'] = [
        'include',
        'exclude',
        'acl',
        'sse',
        'sse-c',
        'sse-c-key',
        'sse-kms-key-id',
        'sse-c-copy-source',
        'sse-c-copy-source-key',
        'storage-class',
        'grants',
        'website-redirect',
        'content-type',
        'cache-control',
        'content-disposition',
        'content-encoding',
        'content-language',
        'expires',
        'source-region',
        'page-size',
        'request-payer',
        'metadata',
        'metadata-directive',
        'expected-size',
        'endpoint-url',
        'output',
        'query',
        'profile',
        'region',
        'version',
        'color',
        'no-sign-request',
        'ca-bundle',
        'cli-read-timeout',
        'cli-connect-timeout',
    ]

    def __init__(self, command, *args, **options):
        """
        Args:
            command (str):
                can be: cp, ls, mv, rm, sync

            *args: positional arguments

            **options: key value options (e.g. profile)
        """
        self.command = command
        self.args = args

        self._known_flags = self.cmd_known_flags.get(self.command, [])
        self._known_keyvals = self.cmd_known_keyvals.get(self.command, [])
        self._known_flags = self._known_flags + [k.replace('-', '_') for k in self._known_flags]
        self._known_keyvals = self._known_keyvals + [k.replace('-', '_') for k in self._known_keyvals]

        # Key / value CLI arguments
        self._keyval_options = {}
        self._flag_options = {}

        self.update(options)

    def update(self, arg=None, /, **options):
        """
        Update key / value options.

        This function is aware of what options need to be flags versus key/values

        So quiet=True will result in `--quiet`, quiet=False will have include no option.
        Likewise profile=foo will result in `--profile foo` and profile=None will include no option.
        """
        if arg is not None:
            options = ub.udict(arg) | options

        for k, v in options.items():
            if k in self._known_flags:
                self._flag_options[k] = v
            elif k in self._known_keyvals:
                self._keyval_options[k] = v
            else:
                # Fallback to key/value
                self._keyval_options[k] = v

    def finalize(self):
        """
        Returns:
            List[str]: commands suitable for passing to POpen
        """
        parts = ['aws', 's3']
        for k, v in self._flag_options.items():
            if v:
                parts.extend(['--' + k.replace('_', '-')])
        for k, v in self._keyval_options.items():
            if v is not None:
                parts.extend(['--' + k.replace('_', '-'), v])
        parts.append(self.command)
        parts.extend(self.args)
        return parts

    def run(self, check=True, shell=False, capture=False, verbose=3):
        """
        Execute the S3 command

        Returns:
            Dict : ubelt cmd info dict
        """
        final_command = self.finalize()
        print('final_command = {}'.format(ub.urepr(final_command, nl=1)))
        run_info = ub.cmd(final_command, verbose=verbose, shell=shell,
                          capture=capture)
        if check:
            run_info.check_returncode()
        return run_info


def ta2_collate_output(aws_base_command, local_region_dir, local_sites_dir,
                       destination_s3_bucket, performer_suffix='KIT'):
    """
    I think this is for putting the final system regions / sites into the place
    that T&E wants them.
    """
    from glob import glob
    from geowatch.utils import util_fsspec
    assert aws_base_command is None, 'unused'

    def _get_suffixed_basename(local_path):
        base, ext = os.path.splitext(os.path.basename(local_path))
        if performer_suffix is not None and performer_suffix != '':
            return "{}_{}{}".format(base, performer_suffix, ext)
        else:
            return "{}{}".format(base, ext)

    for region in glob(join(local_region_dir, '*.geojson')):

        region_s3_outpath = '/'.join((destination_s3_bucket,
                                      'region_models',
                                      _get_suffixed_basename(region)))
        region = util_fsspec.FSPath.coerce(region)
        region_s3_outpath = util_fsspec.FSPath.coerce(region_s3_outpath)
        region.copy(region_s3_outpath)

    for site in glob(join(local_sites_dir, '*.geojson')):
        site_s3_outpath = '/'.join((destination_s3_bucket,
                                    'site_models',
                                    _get_suffixed_basename(site)))
        site = util_fsspec.FSPath.coerce(site)
        site_s3_outpath = util_fsspec.FSPath.coerce(site_s3_outpath)
        site.copy(site_s3_outpath)


def fixup_and_validate_site_and_region_models(region_dpath, site_dpath):
    """
    Read, fix, and validate all site and region models.
    """
    # Validate and fix all outputs
    from geowatch.geoannots import geomodels
    from kwgis.utils import util_gis
    region_infos = list(util_gis.coerce_geojson_datas(region_dpath, format='json'))
    site_infos = list(util_gis.coerce_geojson_datas(site_dpath, format='json'))

    region_models = []
    for region_info in region_infos:
        fpath = region_info['fpath']
        region = geomodels.RegionModel(**region_info['data'])
        region.fixup()
        fpath.write_text(region.dumps(indent='    '))
        region_models.append(region)
        region.validate()

    site_models = []
    for site_info in site_infos:
        fpath = site_info['fpath']
        site = geomodels.SiteModel(**site_info['data'])
        site.fixup()
        fpath.write_text(site.dumps(indent='    '))
        site.validate()
        site_models.append(site)

    check_region_and_site_models_agree(region_models, site_models)


def check_region_and_site_models_agree(region_models, site_models):
    # Check that region / site models are consistent with each other
    all_site_summaries = list(ub.flatten(region.site_summaries() for region in region_models))
    all_site_headers = [site.header for site in site_models]

    sitesum_dups = ub.find_duplicates(sitesum['properties']['site_id'] for sitesum in all_site_summaries)
    assert not sitesum_dups, 'site summaries have duplicate site-ids'

    site_dups = ub.find_duplicates(header['properties']['site_id'] for header in all_site_headers)
    assert not site_dups, 'site models have duplicate site-ids'

    siteid_to_summary = {sitesum['properties']['site_id']: sitesum for sitesum in all_site_summaries}
    siteid_to_header = {header['properties']['site_id']: header for header in all_site_headers}

    site_ids_in_summaries = set(siteid_to_summary)
    site_ids_in_header = set(siteid_to_header)
    if site_ids_in_summaries != site_ids_in_header:
        diff1 = site_ids_in_header - site_ids_in_header
        diff2 = site_ids_in_summaries - site_ids_in_header
        print(f'diff1 = {ub.urepr(diff1, nl=1)}')
        print(f'diff2 = {ub.urepr(diff2, nl=1)}')
        raise Exception('site-ids differe between summaries and site headers')

    # These properties should be the same in both variants
    common_properties = ['start_date', 'end_date', 'status']

    errors = []
    import math
    for site_id in site_ids_in_summaries:
        summary = siteid_to_summary[site_id]
        header = siteid_to_header[site_id]
        summary_props = ub.udict(summary['properties'])
        header_props = ub.udict(header['properties'])

        summary_common_props = summary_props & common_properties
        header_common_props = header_props & common_properties

        score1 = summary_props.get('score', 1.0)
        score2 = header_props.get('score', 1.0)
        has_diff = summary_common_props != header_common_props
        has_diff |= (not math.isclose(score1, score2))

        if has_diff:
            print(f'FOUND INCONSISTENCY IN {site_id}')
            print(f'header_props = {ub.urepr(header_props, nl=1)}')
            print(f'summary_props = {ub.urepr(summary_props, nl=1)}')
            errors.append(site_id)
        # assert summary_props['start_date'] == header_props['start_date']
        # assert summary_props['end_date'] == header_props['end_date']
        # assert summary_props['status'] == header_props['status']

    if errors:
        raise AssertionError(f'There were {len(errors)} / {len(site_ids_in_summaries)} inconsistent sites')
    else:
        print(f'Checked {len(site_ids_in_summaries)} sites which seem consistent')


class NodeStateDebugger:
    """
    Prints information about the current node that is helpful for debugging.

    Use in the smartflow CLI nodes.

    Maintains some internal state to keep things organized.

    Example:
        >>> from geowatch.utils.util_framework import *  # NOQA
        >>> import ubelt as ub
        >>> watch_appdir_dpath = ub.Path.appdir('geowatch')
        >>> self = NodeStateDebugger()
        >>> self.print_environment()
        >>> self.print_current_state(watch_appdir_dpath)
        >>> config = {'foo': 'bar'}
        >>> self.print_local_invocation(config)
    """

    def __init__(self):
        self.current_iteration = 0

    def print_environment(self):
        """
        Print info about what version of the code we are running on
        """
        import os
        import geowatch
        print(' --- <NODE_ENV> --- ')
        print(' * Print current version of the code & environment')
        ub.cmd('git log -n 1', verbose=3, cwd=ub.Path(geowatch.__file__).parent)
        print('geowatch.__version__ = {}'.format(ub.urepr(geowatch.__version__, nl=1)))
        print('geowatch.__file__ = {}'.format(ub.urepr(geowatch.__file__, nl=1)))
        print('os.environ = {}'.format(ub.urepr(dict(os.environ), nl=1)))

        # Check to make sure our times are in sync with amazon servers
        if 0:
            ub.cmd('date -u', verbose=3)
            ub.cmd('curl http://s3.amazonaws.com -v', verbose=3)
        print(' --- </NODE_ENV> --- ')
        # TASK_IMAGE_NAME = os.environ.get('TASK_IMAGE_NAME', None)
        # if TASK_IMAGE_NAME:
        #     self.print_local_invocation()

    def print_local_invocation(self, config=None):
        """
        Attempt to build a string that will allow the user to start stepping
        through a local run of this smartflow step in IPython.

        Args:
            config (scriptconfig.DataConfig):
                the config used to invoke the script
        """
        # Print out a command to help developers debug this image in a
        # local environment.
        # TODO: make this more generic for other people.
        # This is somewhat ill-defined because we can't know which
        # local machine the user will want to run on but here are issues
        # with the current command:
        # * the external code / data is jon-specific,
        # * the location of the mapped ingress directory is arbitrary.
        # * the location of the local .aws directory is usually correct.
        # * Not every image needs runtime=nvidia
        # * pip cache is only necessary if installing new packages, but
        #   location is a reasonable default.
        # * there may be environment variables passed by smartflow that
        #   also need to be passed here, but we dont want to have a huge
        #   command, so there is a tradeoff.
        print('To run in a similar environment locally:')
        TASK_IMAGE_NAME = os.environ.get('TASK_IMAGE_NAME', None)

        from kwutil.slugify_ext import smart_truncate
        import kwutil
        # import uuid
        # unique_suffix = ub.hash_data(uuid.uuid4())[0:8]
        unique_suffix = kwutil.util_time.isoformat(kwutil.util_time.datetime.now(), pathsafe=1)
        if config is None:
            step_name = 'unknown_step'
        else:
            # TODO: ensure pathsafe name utility
            step_name = smart_truncate(config.__class__.__name__, 24, trunc_loc=1, head="_", tail="_")
        temp_location = f'debug_{step_name}_{unique_suffix}'
        create_local_env_command = ub.codeblock(
            fr'''
            # Set these environment variables to reasonable locations on your
            # host machine.
            LOCAL_WORK_DPATH=$HOME/temp/{temp_location}/ingress
            LOCAL_CODE_DPATH=$HOME/code
            LOCAL_DATA_DPATH=$HOME/data

            # Run the docker image
            mkdir -p "$LOCAL_WORK_DPATH"
            cd "$LOCAL_WORK_DPATH"
            docker run \
                --gpus=all \
                --volume "$LOCAL_WORK_DPATH":/tmp/ingress \
                --volume "$LOCAL_CODE_DPATH":/extern_code:ro \
                --volume "$LOCAL_DATA_DPATH":/extern_data:ro \
                --volume "$HOME"/.aws:/root/.aws:ro \
                --volume "$HOME"/.cache/pip:/pip_cache \
                --env AWS_PROFILE=iarpa \
                --env TASK_IMAGE_NAME={TASK_IMAGE_NAME} \
                -it {TASK_IMAGE_NAME} bash
            ''')
        # Do we use --gpus all or --runtime=nvidia
        print()
        print(create_local_env_command)

        if 1:
            helper_text = ub.codeblock(
                '''
                # Point the container repo at the repo on your host
                # system to pull in any updates for testing.
                git remote add host /extern_code/geowatch/.git
                git fetch host
                # may need to do some of:
                # BRANCH=<branch>
                # git reset --hard "host/$BRANCH"
                # git checkout "$BRANCH"
                # git pull "host/$BRANCH"
                ''')
            print()
            print(helper_text)

        # node_modname = 'geowatch.cli.smartflow.run_sc_datagen'
        if config is not None:
            node_modname = config.__class__.__module__
            if node_modname == '__main__':
                # Try to get the real name
                import sys
                node_modname = ub.modpath_to_modname(sys.modules[node_modname].__file__)

            ipython_setup_command = ub.codeblock(
                f'''
                %load_ext autoreload
                %autoreload 2
                from {node_modname} import *
                ''')
            # TODO: some configs which have quotes need to be escaped,
            # otherwise they break here.
            config_text = 'config = ' + ub.urepr(config, nl=1)
            ipython_setup_command = ipython_setup_command + '\n' + config_text
            print()
            print('# In IPython')
            # print(ipython_setup_command)
            wrapped_text = "ipython -i -c 'if 1:\n"
            wrapped_text += ub.indent(ipython_setup_command).replace("'", '"')
            wrapped_text += "\n'"
            print(wrapped_text)

    def print_current_state(self, dpath):
        print(f' --- <NODE_STATE iter={self.current_iteration}> --- ')
        dpath = ub.Path(dpath).resolve()
        # cwd_paths = sorted([p.resolve() for p in dpath.glob('*')])
        # print('cwd_paths = {}'.format(ub.urepr(cwd_paths, nl=1)))
        self.print_directory_contents(dpath)

        print(f' * Print some disk and machine statistics ({self.current_iteration})')
        ub.cmd('df -h', verbose=3)

        from geowatch.utils import util_hardware
        with_units = bool(ub.modname_to_modpath('pint'))
        mem_info = util_hardware.get_mem_info(with_units=with_units)
        print('mem_info = {}'.format(ub.urepr(mem_info, nl=1, align=':')))

        print(f' --- </NODE_STATE iter={self.current_iteration}> --- ')
        self.current_iteration += 1

    def print_directory_contents(self, dpath):
        print(f'* Printing directory contents: {dpath}')
        dpath = ub.Path(dpath).resolve()
        # cwd_paths = sorted([p.resolve() for p in dpath.glob('*')])
        # print('cwd_paths = {}'.format(ub.urepr(cwd_paths, nl=1)))
        if dpath.exists():
            ub.cmd('ls -al', verbose=3, cwd=dpath)
        else:
            print(f'dpath={dpath} does not exist')


class PrintLogger:
    """
    Ducktype a logger
    """
    def info(self, msg, *args, **kwargs):
        print(msg % args)
    def debug(self, msg, *args, **kwargs):
        print(msg % args)
    def error(self, msg, *args, **kwargs):
        print(msg % args)
    def warning(self, msg, *args, **kwargs):
        print(msg % args)
    def critical(self, msg, *args, **kwargs):
        print(msg % args)


def _test_s3_hack():
    """
    An issue that can occur in will manifest as:

    [2023-10-27, 23:28:51 UTC] {pod_manager.py:342} INFO - botocore.exceptions.ClientError: An error occurred (RequestTimeTooSkewed) when calling the PutObject operation: The difference between the request time and the current time is too large.


    botocore.exceptions.ClientError: An error occurred (RequestTimeTooSkewed) when calling the PutObject operation: The difference between the request time and the current time is too large.
    File "/root/code/watch/geowatch/cli/smartflow_egress.py", line 174, in smartflow_egress
    local_path.copy(asset_s3_outpath)
    File "/root/.pyenv/versions/3.11.2/lib/python3.11/site-packages/s3fs/core.py", line 140, in _error_wrapper
    raise err
    PermissionError: The difference between the request time and the current time is too large.
    """
    from geowatch.utils import util_fsspec
    util_fsspec.S3Path._new_fs(profile='iarpa')
    s3_dpath = util_fsspec.S3Path.coerce('s3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v120/batch/kit/CN_C000/2021-08-31/split/mono/products/dummy-test')
    s3_dpath.parent.ls()
    dst_dpath = s3_dpath

    dpath = util_fsspec.LocalPath.appdir('geowatch/fsspec/test-s3-hack/').ensuredir()
    # dst_dpath = (dpath / 'dst')

    src_dpath = (dpath / 'src').ensuredir()
    for i in range(100):
        (src_dpath / f'file_{i:03d}.txt').write_text('hello world' * 100)

    from fsspec.callbacks import TqdmCallback
    callback = TqdmCallback(tqdm_kwargs={"desc": "Your tqdm description"})
    src_dpath.copy(dst_dpath, callback=callback)
    from geowatch.utils import util_framework

    aws_cmd = util_framework.AWS_S3_Command(
        'sync', src_dpath, dst_dpath,
        **dst_dpath.fs.storage_options)
    print(aws_cmd.finalize())
    aws_cmd.run()



import ubelt as ub
import functools
import numbers
from kwcoco.util import dict_proxy2


try:
    from lark import Transformer
except ImportError:
    class Transformer:
        pass


try:
    cache = functools.cache
except AttributeError:
    cache = ub.memoize


# For common constructs see:
# https://github.com/lark-parser/lark/blob/master/lark/grammars/common.lark
RESOLUTION_GRAMMAR_PARTS = ub.codeblock(
    '''
    // Resolution parts of the grammar.
    magnitude: NUMBER

    unit: WORD

    numeric_unit: (magnitude WS* unit)
    implicit_unit: unit

    resolved_unit: numeric_unit | implicit_unit

    %import common.NUMBER
    %import common.WS
    %import common.WORD
    ''')

RESOLVED_UNIT_GRAMMAR = ub.codeblock(
    r'''
    // RESOLVED WINDOW GRAMMAR. Eg. 2GSD
    ?start: resolved_unit
    ''') + '\n' + RESOLUTION_GRAMMAR_PARTS


RESOLVED_SCALAR_GRAMMAR = ub.codeblock(
    r'''
    // RESOLVED WINDOW GRAMMAR. 128 @ 2GSD
    ?start: resolved_scalar

    resolved_scalar: NUMBER WS* "@" WS* resolved_unit

    ''') + '\n' + RESOLUTION_GRAMMAR_PARTS


RESOLVED_WINDOW_GRAMMAR = ub.codeblock(
    r'''
    // RESOLVED WINDOW GRAMMAR. E.g. 128x128 @ 2GSD
    ?start: resolved_window

    window_1d_dim: NUMBER

    window_2d_dim: NUMBER WS* ("x" | ",") WS* NUMBER

    window: window_1d_dim | window_2d_dim

    resolved_window: window WS* "@" WS* resolved_unit
    ''') + '\n' + RESOLUTION_GRAMMAR_PARTS


def _int_or_float(x):
    try:
        return int(x)
    except Exception:
        return float(x)


class ExtendedTransformer(Transformer):
    """
    Enriches the Transformer with parse and parser classmethods which rely on a
    __grammar__ attribute
    """
    __grammar__ = NotImplemented

    @classmethod
    @cache
    def parser(cls):
        # https://github.com/lark-parser/lark/blob/master/docs/_static/lark_cheatsheet.pdf
        import lark
        try:
            import lark_cython
            parser = lark.Lark(cls.__grammar__, start='start', parser='lalr', _plugins=lark_cython.plugins)
        except ImportError:
            parser = lark.Lark(cls.__grammar__, start='start', parser='lalr')
        return parser

    @classmethod
    @cache
    def parse(cls, text):
        """
        Parses the text and transforms the output tree based on
        __grammar__
        """
        parser = cls.parser()
        try:
            tree = parser.parse(text)
        except TypeError:
            if isinstance(text, str) and type(text) is not str:
                # We could be in a case where cython is failing to handle
                # overloaded string types. Try casting to a regular str.
                tree = parser.parse(str(text))
            else:
                raise
        self = cls()
        transformed = self.transform(tree)
        return transformed


class ResolvedTransformer(ExtendedTransformer):
    """
    Base class for resolving a resolution 1D scalar or 2D window.
    """

    def magnitude(self, items):
        d = _int_or_float(items[0].value)
        return d

    def unit(self, items):
        return items[0].value

    def implicit_unit(self, items):
        info = {
            'mag': 1,
            'unit': items[-1],
        }
        return info

    def numeric_unit(self, items):
        info = {
            'mag': items[0],
            'unit': items[-1],
        }
        return info

    def resolved_unit(self, items):
        info = items[0]
        return info


class ResolvedUnitTransformer(ResolvedTransformer):
    """
    Transform for :class:`ResolvedUnit`
    """
    __grammar__ = RESOLVED_UNIT_GRAMMAR


class ResolvedScalarTransformer(ResolvedTransformer):
    """
    Transform for :class:`ResolvedScalar`
    """
    __grammar__ = RESOLVED_SCALAR_GRAMMAR

    def resolved_scalar(self, items):
        info = {
            'scalar': _int_or_float(items[0].value),
            'resolution': items[-1],
        }
        return info


class ResolvedWindowTransformer(ResolvedTransformer):
    """
    Transform for :class:`ResolvedWindow`
    """
    __grammar__ = RESOLVED_WINDOW_GRAMMAR

    def window_1d_dim(self, items):
        d1 = _int_or_float(items[0].value)
        info = (d1, d1)
        return info

    def window_2d_dim(self, items):
        d1 = _int_or_float(items[0].value)
        d2 = _int_or_float(items[-1].value)
        info = (d1, d2)
        return info

    def window(self, items):
        return items[0]

    def resolved_window(self, items):
        info = {
            'window': items[0],
            'resolution': items[-1],
        }
        return info


class Resolved(dict_proxy2.DictProxy2):
    """
    Base class for all resolved objects.
    Must define the ``__transformer__`` attribute.
    """
    __transformer__ = NotImplemented

    @classmethod
    def parse(cls, data):
        if isinstance(data, str):
            text = data
            transformer_cls = cls.__transformer__
            attrs = transformer_cls.parse(text)
            return cls(**attrs)
        else:
            raise TypeError(type(data))

    @classmethod
    def coerce(cls, data):
        if isinstance(data, cls):
            return data
        elif isinstance(data, str):
            return cls.parse(data)
        elif isinstance(data, dict):
            return cls(**data)
        else:
            raise TypeError(type(data))


class ResolvedUnit(Resolved, ub.NiceRepr):
    """
    Holds just the unit information (e.g. X GSD)

    Example:
        >>> from geowatch.utils import util_resolution
        >>> self = util_resolution.ResolvedUnit.parse('8GSD')
        >>> print('self = {}'.format(ub.urepr(self, nl=1, si=1)))
        self = <ResolvedUnit(8 GSD)>
    """
    __transformer__ = ResolvedUnitTransformer

    def __init__(self, mag, unit):
        self.mag = mag
        self.unit = unit
        self._proxy = {
            'mag': mag,
            'unit': unit,
        }

    def __eq__(self, other):
        if self.unit != other.unit:
            raise TypeError(f'incomparable units: {self.unit}, {other.unit}')
        return self.mag == other.mag

    def __nice__(self):
        return (f'{self.mag} {self.unit}')

    @classmethod
    def coerce(cls, data, default_unit=None):
        """
        Example:
            >>> from geowatch.utils.util_resolution import *  # NOQA
            >>> self1 = ResolvedUnit.coerce(8, default_unit='GSD')
            >>> self2 = ResolvedUnit.coerce('8', default_unit='GSD')
            >>> self3 = ResolvedUnit.coerce('8GSD')
            >>> assert self1 == self2
            >>> import pytest
            >>> with pytest.raises(ValueError):
            >>>     ResolvedUnit.coerce(8)

        Example:
            >>> import kwutil
            >>> # Test loading from YAML.
            >>> # https://github.com/lark-parser/lark_cython/issues/36
            >>> from geowatch.utils.util_resolution import ResolvedUnit
            >>> text = kwutil.Yaml.coerce('key: "1 mGSD"')['key']
            >>> ResolvedUnit.coerce(text)
        """
        is_string = isinstance(data, str)
        if is_string:
            # Allow the input to be given as a numeric string
            try:
                mag = _int_or_float(data)
            except Exception:
                ...
            else:
                data = mag
                is_string = False

        if isinstance(data, str):
            self = cls.parse(data)
        elif isinstance(data, numbers.Number):
            if default_unit is None:
                raise ValueError(
                    'must provide a default unit if numberic input is given')
            self = cls(data, default_unit)
        else:
            raise TypeError(type(data))
        return self

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__(self.mag * other, self.unit)
        else:
            raise NotImplementedError

    def __truediv__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__(self.mag / other, self.unit)
        else:
            raise NotImplementedError


class ResolvedScalar(Resolved, ub.NiceRepr):
    """
    Example:
        >>> from geowatch.utils.util_resolution import *  # NOQA
        >>> self1 = ResolvedScalar.parse("128@10GSD")
        >>> self2 = ResolvedScalar.parse("128  @  10  GSD")
        >>> print('self1 = {}'.format(ub.urepr(self1, sv=1, nl=1)))
        >>> print('self2 = {}'.format(ub.urepr(self2, sv=1, nl=1)))
        self1 = <ResolvedScalar(128 @ 10 GSD)>
        self2 = <ResolvedScalar(128 @ 10 GSD)>
    """
    __transformer__ = ResolvedScalarTransformer

    def __init__(self, scalar, resolution):
        self.scalar = scalar
        self.resolution = ResolvedUnit(**resolution)
        self._proxy = {
            'scalar': scalar,
            'resolution': resolution,
        }

    def __nice__(self):
        return (f'{self.scalar} @ {self.resolution.__nice__()}')

    def at_resolution(self, new_resolution):
        '''
        Update the resolution

        Args:
            new_resolution (dict | ResolvedUnit):
                new base resolution unit to use.

        Returns:
            ResolvedScalar:
                The same scalar but in terms of the new resolution.

        Example:
            >>> new_resolution = {'mag': 1, 'unit': 'GSD'}
            >>> self = ResolvedScalar.parse("128@10GSD")
            >>> print(self.at_resolution(new_resolution))
            >>> print(self.at_resolution({'mag': 20, 'unit': 'GSD'}))
            <ResolvedScalar(1280.0 @ 1 GSD)>
            <ResolvedScalar(64.0 @ 20 GSD)>
        '''
        scale_factor = self.resolution['mag'] / new_resolution['mag']
        new = self.__class__(self.scalar * scale_factor, new_resolution)
        return new


class ResolvedWindow(Resolved, ub.NiceRepr):
    """
    Parse a window size at a particular resolution

    Example:
        >>> from geowatch.utils.util_resolution import *  # NOQA
        >>> data = "128x128@10GSD"
        >>> self1 = ResolvedWindow.parse(data)
        >>> self2 = ResolvedWindow.parse("128  ,  128  @  10  GSD")
        >>> self3 = ResolvedWindow.parse("128@10GSD")
        >>> print('self1 = {}'.format(ub.urepr(self1, nl=1, sv=1)))
        >>> print('self2 = {}'.format(ub.urepr(self2, nl=1, sv=1)))
        >>> print('self3 = {}'.format(ub.urepr(self3, nl=1, sv=1)))
        self1 = <ResolvedWindow((128, 128) @ 10 GSD)>
        self2 = <ResolvedWindow((128, 128) @ 10 GSD)>
        self3 = <ResolvedWindow((128, 128) @ 10 GSD)>
    """
    __transformer__ = ResolvedWindowTransformer

    def __init__(self, window, resolution):
        self.window = window
        self.resolution = ResolvedUnit(**resolution)
        self._proxy = {
            'window': window,
            'resolution': resolution,
        }

    def at_resolution(self, new_resolution):
        """
        Update the resolution

        Args:
            new_resolution (dict | ResolvedUnit):
                new base resolution unit to use.

        Returns:
            ResolvedWindow:
                The same window but in terms of the new resolution.

        Example:
            >>> from geowatch.utils.util_resolution import *  # NOQA
            >>> new_resolution = {'mag': 1, 'unit': 'GSD'}
            >>> self = ResolvedWindow.parse("128x64@10GSD")
            >>> print(self.at_resolution(new_resolution))
            >>> print(self.at_resolution({'mag': 20, 'unit': 'GSD'}))
            <ResolvedWindow((1280.0, 640.0) @ 1 GSD)>
            <ResolvedWindow((64.0, 32.0) @ 20 GSD)>
        """
        scale_factor = self.resolution['mag'] / new_resolution['mag']
        w, h = self.window
        new_window = (w * scale_factor, h * scale_factor)
        new = self.__class__(new_window, new_resolution)
        return new

    def __nice__(self):
        return (f'{self.window} @ {self.resolution.__nice__()}')

    # TODO:
    # More flexible inputs
    # @classmethod
    # def coerce(cls, data, default_unit=None):
    #     """
    #     Example:
    #         >>> from geowatch.utils.util_resolution import *  # NOQA
    #         >>> windows = []
    #         >>> windows.append(ResolvedWindow.coerce("128@10GSD"))
    #         >>> windows.append(ResolvedWindow.coerce(128, default_unit='10mGSD'))
    #         >>> windows.append(ResolvedWindow.coerce([128, 128], default_unit='10mGSD'))
    #         >>> windows.append(ResolvedWindow.coerce(1280, default_unit='mGSD'))
    #         >>> windows.append(ResolvedWindow.coerce('1280', default_unit='mGSD'))
    #         >>> print(f'windows = {ub.urepr(windows, nl=1)}')
    #     """
    #     is_string = isinstance(data, str)
    #     if is_string:
    #         # Allow the input to be given as a numeric string
    #         try:
    #             mag = _int_or_float(data)
    #         except Exception:
    #             ...
    #         else:
    #             data = mag
    #             is_string = False

    #     if isinstance(data, str):
    #         self = cls.parse(data)

    #     if ub.iterable(data):
    #         self = cls.parse(data)

    #     elif isinstance(data, numbers.Number):
    #         if default_unit is None:
    #             raise ValueError(
    #                 'must provide a default unit if numberic input is given')
    #         default_unit = ResolvedUnit.coerce(default_unit)
    #         self = cls(data, default_unit)
    #     else:
    #         raise TypeError(type(data))
    #     return self



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers and conversions for third party pycocotools.

This is derived from code in the Tensorflow Object Detection API:
https://github.com/tensorflow/models/tree/master/research/object_detection

Huang et. al. "Speed/accuracy trade-offs for modern convolutional object
detectors" CVPR 2017.
"""

from typing import Any, Collection, Dict, List, Optional, Union

import numpy as np
from pycocotools import mask

COCO_METRIC_NAMES_AND_INDEX = (
    ('Precision/mAP', 0),
    ('Precision/mAP@.50IOU', 1),
    ('Precision/mAP@.75IOU', 2),
    ('Precision/mAP (small)', 3),
    ('Precision/mAP (medium)', 4),
    ('Precision/mAP (large)', 5),
    ('Recall/AR@1', 6),
    ('Recall/AR@10', 7),
    ('Recall/AR@100', 8),
    ('Recall/AR@100 (small)', 9),
    ('Recall/AR@100 (medium)', 10),
    ('Recall/AR@100 (large)', 11)
)


def _ConvertBoxToCOCOFormat(box: np.ndarray) -> List[float]:
    """Converts a box in [ymin, xmin, ymax, xmax] format to COCO format.

    This is a utility function for converting from our internal
    [ymin, xmin, ymax, xmax] convention to the convention used by the COCO API
    i.e., [xmin, ymin, width, height].

    Args:
      box: a [ymin, xmin, ymax, xmax] numpy array

    Returns:
      a list of floats representing [xmin, ymin, width, height]
    """
    return [float(box[1]), float(box[0]), float(box[3] - box[1]),
            float(box[2] - box[0])]


def ExportSingleImageGroundtruthToCoco(
        image_id: Union[int, str],
        next_annotation_id: int,
        category_id_set: Collection[int],
        groundtruth_boxes: np.ndarray,
        groundtruth_classes: np.ndarray,
        groundtruth_masks: np.ndarray,
        groundtruth_is_crowd: Optional[np.ndarray] = None) -> List[Dict[str, Any]]:
    """Exports groundtruth of a single image to COCO format.

    This function converts groundtruth detection annotations represented as numpy
    arrays to dictionaries that can be ingested by the COCO evaluation API. Note
    that the image_ids provided here must match the ones given to
    ExportSingleImageDetectionsToCoco. We assume that boxes and classes are in
    correspondence - that is: groundtruth_boxes[i, :], and
    groundtruth_classes[i] are associated with the same groundtruth annotation.

    In the exported result, "area" fields are always set to the foregorund area of
    the mask.

    Args:
      image_id: a unique image identifier either of type integer or string.
      next_annotation_id: integer specifying the first id to use for the
        groundtruth annotations. All annotations are assigned a continuous integer
        id starting from this value.
      category_id_set: A set of valid class ids. Groundtruth with classes not in
        category_id_set are dropped.
      groundtruth_boxes: numpy array (float32) with shape [num_gt_boxes, 4]
      groundtruth_classes: numpy array (int) with shape [num_gt_boxes]
      groundtruth_masks: uint8 numpy array of shape [num_detections, image_height,
        image_width] containing detection_masks.
      groundtruth_is_crowd: optional numpy array (int) with shape [num_gt_boxes]
        indicating whether groundtruth boxes are crowd.

    Returns:
      a list of groundtruth annotations for a single image in the COCO format.

    Raises:
      ValueError: if (1) groundtruth_boxes and groundtruth_classes do not have the
        right lengths or (2) if each of the elements inside these lists do not
        have the correct shapes or (3) if image_ids are not integers
    """

    if len(groundtruth_classes.shape) != 1:
        raise ValueError('groundtruth_classes is '
                         'expected to be of rank 1.')
    if len(groundtruth_boxes.shape) != 2:
        raise ValueError('groundtruth_boxes is expected to be of '
                         'rank 2.')
    if groundtruth_boxes.shape[1] != 4:
        raise ValueError('groundtruth_boxes should have '
                         'shape[1] == 4.')
    num_boxes = groundtruth_classes.shape[0]
    if num_boxes != groundtruth_boxes.shape[0]:
        raise ValueError('Corresponding entries in groundtruth_classes, '
                         'and groundtruth_boxes should have '
                         'compatible shapes (i.e., agree on the 0th dimension).'
                         'Classes shape: %d. Boxes shape: %d. Image ID: %s' % (
                             groundtruth_classes.shape[0],
                             groundtruth_boxes.shape[0], image_id))
    has_is_crowd = groundtruth_is_crowd is not None
    if has_is_crowd and len(groundtruth_is_crowd.shape) != 1:
        raise ValueError('groundtruth_is_crowd is expected to be of rank 1.')
    groundtruth_list = []
    for i in range(num_boxes):
        if groundtruth_classes[i] in category_id_set:
            iscrowd = groundtruth_is_crowd[i] if has_is_crowd else 0
            segment = mask.encode(np.asfortranarray(groundtruth_masks[i]))
            area = mask.area(segment)
            export_dict = {
                'id': next_annotation_id + i,
                'image_id': image_id,
                'category_id': int(groundtruth_classes[i]),
                'bbox': list(_ConvertBoxToCOCOFormat(groundtruth_boxes[i, :])),
                'segmentation': segment,
                'area': area,
                'iscrowd': iscrowd
            }

            groundtruth_list.append(export_dict)
    return groundtruth_list


def ExportSingleImageDetectionMasksToCoco(
        image_id: Union[int, str], category_id_set: Collection[int],
        detection_masks: np.ndarray, detection_scores: np.ndarray,
        detection_classes: np.ndarray) -> List[Dict[str, Any]]:
    """Exports detection masks of a single image to COCO format.

    This function converts detections represented as numpy arrays to dictionaries
    that can be ingested by the COCO evaluation API. We assume that
    detection_masks, detection_scores, and detection_classes are in correspondence
    - that is: detection_masks[i, :], detection_classes[i] and detection_scores[i]
      are associated with the same annotation.

    Args:
      image_id: unique image identifier either of type integer or string.
      category_id_set: A set of valid class ids. Detections with classes not in
        category_id_set are dropped.
      detection_masks: uint8 numpy array of shape [num_detections, image_height,
        image_width] containing detection_masks.
      detection_scores: float numpy array of shape [num_detections] containing
        scores for detection masks.
      detection_classes: integer numpy array of shape [num_detections] containing
        the classes for detection masks.

    Returns:
      a list of detection mask annotations for a single image in the COCO format.

    Raises:
      ValueError: if (1) detection_masks, detection_scores and detection_classes
        do not have the right lengths or (2) if each of the elements inside these
        lists do not have the correct shapes or (3) if image_ids are not integers.
    """

    if len(detection_classes.shape) != 1 or len(detection_scores.shape) != 1:
        raise ValueError('All entries in detection_classes and detection_scores'
                         'expected to be of rank 1.')
    num_boxes = detection_classes.shape[0]
    if not num_boxes == len(detection_masks) == detection_scores.shape[0]:
        raise ValueError('Corresponding entries in detection_classes, '
                         'detection_scores and detection_masks should have '
                         'compatible lengths and shapes '
                         'Classes length: %d.  Masks length: %d. '
                         'Scores length: %d' % (
                             detection_classes.shape[0], len(detection_masks),
                             detection_scores.shape[0]
                         ))
    detections_list = []
    for i in range(num_boxes):
        if detection_classes[i] in category_id_set:
            detections_list.append({
                'image_id': image_id,
                'category_id': int(detection_classes[i]),
                'segmentation': mask.encode(np.asfortranarray(detection_masks[i])),
                'score': float(detection_scores[i])
            })
    return detections_list



"""
This submodule contains tools relating to:

    * the IARPA region and site model geojson formats.
    * geographic kwcoco extensions

At the time of creation we have many various tools scattered across the repo to
deal with these, but we should work to consolidate the ones pertaining strictly
to these specific types of geojsons into this submodule.


Location of existing tools that should likely be moved:

    * ../demo/metrics_demo/__init__.py :: various
    * ../cli/run_tracker.py :: various
    * ../cli/reproject_annotations.py :: various
    * ../cli/validate_annotation_schemas.py :: various
    * ../cli/extend_sc_sites.py :: various
    * ../cli/crop_sites_to_regions.py :: various
    * ../cli/cluster_sites.py :: various
    * ../utils/util_framework.py :: determine_region_id

    * ~/code/watch/dev/poc/make_region_from_sitemodel.py


Related tool that should NOT be moved are related to general geojson:

    * ../utils/util_gis.py :: coerce_geojson_datas
"""


__mkinit__ = """
mkinit ~/code/geowatch/geowatch/geoannots/__init__.py --lazy --diff

TEST:
    python -c "from geowatch import geoannots"
    EAGER_IMPORT_MODULES=geowatch python -c "from geowatch import geoannots"
"""
# import lazy_loader

__submodules__ = {
    'geococo_objects': [],
    'geomodels': [],
}


def lazy_import(module_name, submodules, submod_attrs, eager='auto'):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    eager_import_flag = False
    if eager == 'auto':
        eager_import_text = os.environ.get('EAGER_IMPORT', '')
        if eager_import_text:
            eager_import_text_ = eager_import_text.lower()
            if eager_import_text_ in {'true', '1', 'on', 'yes'}:
                eager_import_flag = True

        eager_import_module_text = os.environ.get('EAGER_IMPORT_MODULES', '')
        if eager_import_module_text:
            if __name__.lower() in eager_import_module_text.lower():
                eager_import_flag = True
    else:
        eager_import_flag = eager

    if eager_import_flag:
        for name in submodules:
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'geococo_objects',
        'geomodels',
    },
    submod_attrs={},
)


def __dir__():
    return __all__

__all__ = ['geococo_objects', 'geomodels']



def kwcoco_footprint_example():
    import watch
    import kwimage
    import kwcoco

    USE_TOYDATA = 0
    if USE_TOYDATA:
        coco_dset = watch.coerce_kwcoco()
    else:
        dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        kwcoco_fpath = dvc_dpath / 'Drop4-BAS/data_vali.kwcoco.json'
        coco_dset = kwcoco.CocoDataset(kwcoco_fpath)

    # Loop through all videos / images like:
    for video_id in coco_dset.videos():
        for image_id in coco_dset.images(video_id=video_id):

            # Construct a CocoImage object, which is a helper to make it eaiser
            # to query pixel and warp / transform information (note annotation
            # access could be easier).
            coco_img : kwcoco.CocoImage = coco_dset.coco_image(image_id)

            # The coco image doesn't give access to annotations yet, but the
            # regular coco API does, so we can lookup an "annotations" object
            # representing all annotations in the image.
            annots = coco_dset.annots(gid=coco_img.img['id'])

            # Accessing the "detections" attribute returns a kwimage.Detections
            # object in IMAGE_SPACE (i.e. aligned to the pixels of the highest
            # resolution asset / auxiliary band registered in the image).
            dets : kwimage.Detections = annots.detections
            # The detections object contains boxes, segmentation, and category
            # information.
            dets.data['boxes']
            dets.data['segmentations']
            dets.data['class_idxs']
            # The class indexes are are indexes into these object "classes"
            dets.meta['classes']

            # However because the detections are returned in IMAGE space.  we
            # may need to warp them into VIDEO space. Fortunately this is easy.
            # The CocoImage object provides this transform matrix.
            warp_vid_from_img : kwimage.Affine = coco_img.warp_vid_from_img

            # The kwimage "warp" method makes it easy jointly transform all
            # annotation objects.
            vidspace_dets = dets.warp(warp_vid_from_img)

            # The underlying warped videospace boxes can be accessed as such:
            boxes : kwimage.Boxes = vidspace_dets.boxes

            # But we can also access polygons
            polys : kwimage.SegmentationList = vidspace_dets.data['segmentations']

            # Bounding boxes of polygons can be accessed by taking the bounds
            # of the polygons. (Which may be more accurate because transforming
            # polygons is lossless whereas transforming boxes is lossy)
            box_accum = []
            for poly in polys:
                box_accum.append(poly.to_boxes())
            boxes_v2 = kwimage.Boxes.concatenate(box_accum)



#!/usr/bin/env python3
# FIXME: broken
import scriptconfig as scfg


class FitCLI(scfg.DataConfig):
    """
    Does not work from geowatch CLI yet. See help.
    Use ``python -m geowatch.tasks.fusion.fit_lightning fit`` instead for now. See

    ..code:: bash

        python -m geowatch.tasks.fusion.fit_lightning fit --help

    Execute fusion fit
    """
    __command__ = 'fit'
    __alias__ = ['fusion_fit']

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        """
        from geowatch.tasks.fusion import fit_lightning
        return fit_lightning.main(cmdline=cmdline, **kwargs)


__cli__ = FitCLI
main = __cli__.main

if __name__ == '__main__':
    main()



import kornia  # NOQA

import torch

from kornia.filters.kernels import normalize_kernel2d

from .kernels import get_spatial_gradient_kernel2d


class SpatialGradient(torch.nn.Module):
    r"""Computes the first order image derivative in both x and y using a Sobel or Scharr
    operator.

    Return:
        torch.Tensor: the sobel edges of the input feature map.

    Shape:
        - Input: :math:`(B, C, H, W)`
        - Output: :math:`(B, C, 2, H, W)`

    Examples:
        input = torch.rand(1, 3, 4, 4)
        output = kornia.filters.SpatialGradient()(input)  # 1x3x2x4x4
    """

    def __init__(self,
                 mode: str = 'sobel',
                 order: int = 1,
                 normalized: bool = True,
                 coord: str = "xy",
                 device: str = "cpu",
                 dtype: torch.dtype = torch.float) -> None:
        super(SpatialGradient, self).__init__()
        self.normalized: bool = normalized
        self.order: int = order
        self.mode: str = mode
        self.kernel: torch.Tensor = get_spatial_gradient_kernel2d(mode, order, coord)
        if self.normalized:
            self.kernel = normalize_kernel2d(self.kernel)
        # Pad with "replicate for spatial dims, but with zeros for channel
        self.spatial_pad = [self.kernel.size(1) // 2,
                            self.kernel.size(1) // 2,
                            self.kernel.size(2) // 2,
                            self.kernel.size(2) // 2]
        # Prepare kernel
        self.kernel: torch.Tensor = self.kernel.to(device).to(dtype).detach()
        self.kernel: torch.Tensor = self.kernel.unsqueeze(1).unsqueeze(1)
        self.kernel: torch.Tensor = self.kernel.flip(-3)
        return

    def __repr__(self) -> str:
        return self.__class__.__name__ + '('\
            'order=' + str(self.order) + ', ' + \
            'normalized=' + str(self.normalized) + ', ' + \
            'mode=' + self.mode + ')'

    def forward(self, inp: torch.Tensor) -> torch.Tensor:  # type: ignore
        if not torch.is_tensor(inp):
            raise TypeError("Input type is not a torch.Tensor. Got {}"
                            .format(type(inp)))
        if not len(inp.shape) == 4:
            raise ValueError("Invalid input shape, we expect BxCxHxW. Got: {}"
                             .format(inp.shape))
        # prepare kernel
        b, c, h, w = inp.shape

        # convolve inp tensor with sobel kernel
        out_channels: int = 3 if self.order == 2 else 2
        padded_inp: torch.Tensor = torch.nn.functional.pad(inp.reshape(b * c, 1, h, w),
                                                           self.spatial_pad, 'replicate')[:, :, None]
        return torch.nn.functional.conv3d(padded_inp, self.kernel, padding=0).view(b, c, out_channels, h, w)



"""
Demo how to efficiently load an image at a specified resolution.
"""
import watch
dset = watch.coerce_kwcoco('watch-msi', geodata=True)
coco_img = dset.images().coco_images[0]

delayed = coco_img.imdelay(
    space='video', resolution='0.1 GSD', RESOLUTION_KEY='target_gsd')

image = delayed.finalize()



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


#https://github.com/clcarwin/focal_loss_pytorch/blob/master/focalloss.py
class FocalLoss(nn.Module):
    def __init__(self, gamma=0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list):
            self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        if input.dim() > 2:
            input = input.view(input.size(0), input.size(1), -1)  # N,C,H,W => N,C,H*W
            input = input.transpose(1, 2)    # N,C,H*W => N,H*W,C
            input = input.contiguous().view(-1, input.size(2))   # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)
        logpt = F.log_softmax(input)
        logpt = logpt.gather(1, target)
        logpt = logpt.view(-1)
        pt = Variable(logpt.data.exp())

        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.data.view(-1))
            logpt = logpt * Variable(at)

        loss = -1 * (1 - pt)**self.gamma * logpt
        if self.size_average:
            return loss.mean()
        else:
            return loss.sum()


class BinaryFocalLoss(nn.Module):
    def __init__(self, weight=None, gamma=2., alpha=1, reduction='mean'):
        nn.Module.__init__(self)
        self.weight = weight
        self.gamma = gamma
        self.reduction = reduction
        self.alpha = alpha

    def forward(self, inputs, targets):
        inputs = inputs.float()
        targets = targets.float()
        BCE_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction=self.reduction)
        pt = torch.exp(-BCE_loss)  # prevents nans when probability 0
        F_loss = self.alpha * (1 - pt) ** self.gamma * BCE_loss
        return F_loss.mean()



"""
Utilities for dictionaries where dots in keys represent nestings
"""
import ubelt as ub
import pygtrie


class DotDict(ub.UDict):
    """
    I'm sure this data structure exists on pypi.
    This should be replaced with that if we find it.

    SeeAlso:
        DotDictDataFrame

    Example:
        >>> from geowatch.utils.util_dotdict import *  # NOQA
        >>> self = DotDict({
        >>>     'proc1.param1': 1,
        >>>     'proc1.param2': 2,
        >>>     'proc2.param1': 3,
        >>>     'proc2.param2': 4,
        >>>     'proc3.param1': 5,
        >>>     'proc3.param2': 6,
        >>>     'proc4.part1.param1': 7,
        >>>     'proc4.part1.param2': 8,
        >>>     'proc4.part2.param2': 9,
        >>>     'proc4.part2.param2': 10,
        >>> })
        >>> self.get('proc1')
        >>> self.prefix_get('proc4')
        >>> 'proc1' in self

        >>> nested = self.to_nested()
        >>> recon = DotDict.from_nested(nested)
        >>> assert nested != self
        >>> assert recon == self
    """

    def __init__(self, /, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Tries work well with prefix stuff, but they may be too complex for
        # what we really need to do here.
        self._trie_cache = {}

    @classmethod
    def from_nested(cls, data):
        """
        Args:
            data (Dict):
                nested data
        """
        flat = cls()
        walker = ub.IndexableWalker(data, list_cls=tuple())
        for path, value in walker:
            if not isinstance(value, dict):
                spath = list(map(str, path))
                key = '.'.join(spath)
                flat[key] = value
        return flat

    def to_nested(self):
        """
        Converts this flat DotDict into a nested representation.  I.e. keys are
        broken using the "." separtor, with each separator becoming a new
        nesting level.

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict(**{
            >>>     'foo.bar.baz': 1,
            >>>     'foo.bar.biz': 1,
            >>>     'foo.spam': 1,
            >>>     'eggs.spam': 1,
            >>> })
            >>> nested = self.to_nested()
            >>> print(f'nested = {ub.urepr(nested, nl=2)}')
            nested = {
                'foo': {
                    'bar': {'baz': 1, 'biz': 1},
                    'spam': 1,
                },
                'eggs': {
                    'spam': 1,
                },
            }
        """
        auto = ub.AutoDict()
        walker = ub.IndexableWalker(auto)
        d = self
        for k, v in d.items():
            path = k.split('.')
            walker[path] = v
        return auto.to_dict()

    def to_nested_keys(self):
        """
        Converts this flat DotDict into a nested key representation.
        The difference between this and to_nested is that the leafs are
        sets of keys whereas the leafs in DotDict are dicts

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict(**{
            >>>     'foo.bar.baz': 1,
            >>>     'foo.bar.biz': 1,
            >>>     'foo.spam': 1,
            >>>     'eggs.spam': 1,
            >>> })
            >>> nested = self.to_nested_keys()
            >>> print(f'nested = {ub.urepr(nested, nl=2)}')
            nested = {
                'foo': {
                    'bar': {'baz': 'foo.bar.baz', 'biz': 'foo.bar.biz'},
                    'spam': 'foo.spam',
                },
                'eggs': {
                    'spam': 'eggs.spam',
                },
            }
        """
        auto = ub.AutoDict()
        walker = ub.IndexableWalker(auto)
        for k in self:
            path = k.split('.')
            walker[path] = k
        # print(ub.urepr(auto))
        return auto.to_dict()

    @property
    def _prefix_trie(self):
        if self._trie_cache.get('prefix_trie', None) is None:
            _trie_data = ub.dzip(self.keys(), self.keys())
            _trie = pygtrie.StringTrie(_trie_data, separator='.')
            self._trie_cache['prefix_trie'] = _trie
        return self._trie_cache['prefix_trie']

    @property
    def _suffix_trie(self):
        if 'suffix_trie' not in self._trie_cache:
            reversed_keys = {
                '.'.join(reversed(k.split('.'))): k
                for k in self.keys()
            }
            _trie = pygtrie.StringTrie(reversed_keys, separator='.')
            self._trie_cache['suffix_trie'] = _trie
        return self._trie_cache['suffix_trie']

    def suffix_get(self, suffix, default=ub.NoParam, backend='trie'):
        """
        Retrieve all key-value pairs whose keys end with a given dot-suffix.

        Args:
            suffix (str): dot-separated suffix string
            default: fallback if no matches found
            backend (str): 'trie' or 'loop'

        Returns:
            DotDict

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict({
            >>>     'a.b.c': 1,
            >>>     'x.b.c': 2,
            >>>     'z.y': 3,
            >>> })
            >>> self.suffix_get('b.c')
            {'a.b.c': 1, 'x.b.c': 2}
        """
        if backend == 'loop':
            matches = DotDict({
                k: v for k, v in self.items()
                if k.endswith('.' + suffix) or k == suffix
            })
        elif backend == 'trie':
            rev_suffix = '.'.join(reversed(suffix.split('.')))
            try:
                matches = DotDict({
                    k: self[k] for k in self._suffix_trie.values(rev_suffix)
                })
            except KeyError:
                if default is not ub.NoParam:
                    return default
                raise
        else:
            raise ValueError(f'Unknown backend={backend}')

        if not matches and default is not ub.NoParam:
            return default
        return matches

    def prefix_get(self, key, default=ub.NoParam):
        """
        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict(**{
            >>>     'foo.bar.baz': 1,
            >>>     'foo.bar.biz': 1,
            >>>     'foo.spam': 1,
            >>>     'eggs.spam': 1,
            >>> })
            >>> self.prefix_get('foo')
            {'bar.baz': 1, 'bar.biz': 1, 'spam': 1}
        """
        try:
            suffix_dict = DotDict()
            full_keys = self._prefix_trie.values(key)
        except KeyError:
            if default is not ub.NoParam:
                return default
            else:
                raise
        else:
            for full_key in full_keys:
                sub_key = full_key[len(key) + 1:]
                suffix_dict[sub_key] = self[full_key]
            return suffix_dict

    def suffix_subdict(self, suffixes, backend='trie'):
        """
        Filter DotDict to only contain keys ending with any given suffixes.

        Args:
            suffixes (List[str]): list of dot-suffixes
            backend (str): 'trie' or 'loop'

        Returns:
            DotDict

        References:
            https://chatgpt.com/c/6841a161-2cd4-8002-ad9b-5593f5a2d70c

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict({
            >>>     'proc1.param1': 1,
            >>>     'proc2.param1': 2,
            >>>     'proc3.param2': 3,
            >>>     'proc4.part1.param1': 4,
            >>>     'proc4.part2.param2': 5,
            >>> })
            >>> new = self.suffix_subdict(['param1', 'part2.param2'])
            >>> print(f'new = {ub.urepr(new, nl=1, sort=1)}')
            new = {
                'proc1.param1': 1,
                'proc2.param1': 2,
                'proc4.part1.param1': 4,
                'proc4.part2.param2': 5,
            }
        """
        if backend == 'loop':
            result = {
                k: v for k, v in self.items()
                if any(k.endswith('.' + suf) or k == suf for suf in suffixes)
            }
        elif backend == 'trie':
            reversed_trie = self._suffix_trie
            result_keys = set()
            for suf in suffixes:
                rev_suf = '.'.join(reversed(suf.split('.')))
                result_keys.update(reversed_trie.values(rev_suf))
            result = {k: self[k] for k in result_keys}
        else:
            raise ValueError(f'Unknown backend={backend}')
        return self.__class__(result)

    def prefix_subdict(self, prefixes, backend='trie'):
        """
        Filter DotDict to only contain keys starting with any given prefixes.

        Args:
            prefixes (List[str]): list of dot-prefixes
            backend (str): 'trie' or 'loop'

        Returns:
            DotDict

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict({
            >>>     'proc1.param1': 1,
            >>>     'proc1.param2': 2,
            >>>     'proc2.param1': 3,
            >>>     'proc3.param2': 4,
            >>>     'proc4.part1.param1': 5,
            >>>     'proc4.part2.param2': 6,
            >>> })
            >>> new = self.prefix_subdict(['proc1', 'proc4.part1'])
            >>> print(f'new = {ub.urepr(new, nl=1, sort=1)}')
            new = {
                'proc1.param1': 1,
                'proc1.param2': 2,
                'proc4.part1.param1': 5,
            }
        """
        if backend == 'loop':
            result = {
                k: v for k, v in self.items()
                if any(k.startswith(pref + '.') or k == pref for pref in prefixes)
            }
        elif backend == 'trie':
            trie = self._prefix_trie
            result_keys = set()
            for pref in prefixes:
                try:
                    result_keys.update(trie.values(pref))
                except KeyError:
                    pass  # It's okay if a prefix has no matches
            result = {k: self[k] for k in result_keys}
        else:
            raise ValueError(f'Unknown backend={backend}')
        return self.__class__(result)

    def add_prefix(self, prefix):
        """
        Adds a prefix to all items
        """
        new = self.__class__([(prefix + '.' + k, v) for k, v in self.items()])
        return new

    def insert_prefix(self, prefix, index):
        """
        Adds a prefix to all items

        Args:
            prefix (str): prefix to insert
            index (int): the depth to insert the new param

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict({
            >>>     'proc1.param1': 1,
            >>>     'proc1.param2': 2,
            >>>     'proc2.param1': 3,
            >>>     'proc4.part1.param2': 8,
            >>>     'proc4.part2.param2': 9,
            >>>     'proc4.part2.param2': 10,
            >>> })
            >>> new = self.insert_prefix('foo', index=1)
            >>> print('self = {}'.format(ub.urepr(self, nl=1)))
            >>> print('new = {}'.format(ub.urepr(new, nl=1)))
        """
        def _generate_new_items():
            sep = '.'
            for k, v in self.items():
                path = k.split(sep)
                path.insert(index, prefix)
                k2 = sep.join(path)
                yield k2, v
        new = self.__class__(_generate_new_items())
        return new

    def query_keys(self, col):
        """
        Finds columns where one level has this key

        Example:
            >>> from geowatch.utils.util_dotdict import *  # NOQA
            >>> self = DotDict({
            >>>     'proc1.param1': 1,
            >>>     'proc1.param2': 2,
            >>>     'proc2.param1': 3,
            >>>     'proc4.part1.param2': 8,
            >>>     'proc4.part2.param2': 9,
            >>>     'proc4.part2.param2': 10,
            >>> })
            >>> list(self.query_keys('param1'))

        Ignore:
            could use _trie_iteritems
            trie = self._prefix_trie
        """
        for key in self.keys():
            if col in set(key.split('.')):
                yield key

    def print_graph(self):
        explore_nested_dict(self)

    # def __contains__(self, key):
    #     if super().__contains__(key):
    #         return True
    #     else:
    #         subkeys = []
    #         subkeys.extend(self._prefix_trie.values(key))
    #         return bool(subkeys)

    # def get(self, key, default=ub.NoParam):
    #     if default is ub.NoParam:
    #         return self[key]
    #     else:
    #         try:
    #             return self[key]
    #         except KeyError:
    #             return default

    # def __getitem__(self, key):
    #     try:
    #         return super().__getitem__(key)
    #     except KeyError:
    #         subkeys = []
    #         subkeys.extend(self._prefix_trie.values(key))
    #         return self.__class__([(k, self[k]) for k in subkeys])


def dotdict_to_nested(d):
    return DotDict.to_nested(d)


def dotkeys_to_nested(keys):
    """
    Args:
        keys (List[str]): a list of dotted key names
    """
    # this is abusing duck typing
    return DotDict.to_nested_keys(keys)


def indexable_to_graph(data):
    import networkx as nx
    graph = nx.DiGraph()
    walker = ub.IndexableWalker(data)
    for path, value in walker:
        spath = list(map(str, path))
        key = '.'.join(spath)
        graph.add_node(key)
        label = spath[-1]
        if not isinstance(value, walker.indexable_cls):
            label = f'{label} : {type(value).__name__} = {value}'

        graph.nodes[key].update({
            'path': path,
            'value': value,
            'label': label,
        })
        if len(path) > 1:
            parent_key = '.'.join(spath[:-1])
            graph.add_edge(parent_key, key)
    return graph


def explore_nested_dict(data):
    """
    TODO: some sort of textual interface
    """
    graph = indexable_to_graph(data)

    from cmd_queue.util.util_networkx import write_network_text
    import rich
    write_network_text(graph, path=rich.print, end='')



"""Code related to optional dependencies."""

import inspect
import os
from contextlib import contextmanager
from importlib.util import find_spec
from typing import Optional

__all__ = [
    "get_config_read_mode",
    "set_config_read_mode",
    "set_docstring_parse_options",
]


typing_extensions_support = find_spec("typing_extensions") is not None
typeshed_client_support = find_spec("typeshed_client") is not None
jsonschema_support = find_spec("jsonschema") is not None
jsonnet_support = find_spec("_jsonnet") is not None
url_support = find_spec("requests") is not None
docstring_parser_support = find_spec("docstring_parser") is not None
fsspec_support = find_spec("fsspec") is not None
ruyaml_support = find_spec("ruyaml") is not None
omegaconf_support = find_spec("omegaconf") is not None
reconplogger_support = find_spec("reconplogger") is not None
attrs_support = find_spec("attrs") is not None

_config_read_mode = "fr"
_docstring_parse_options = {
    "style": None,
    "attribute_docstrings": False,
}


def typing_extensions_import(name):
    if typing_extensions_support:
        return getattr(__import__("typing_extensions"), name, False)
    else:
        return getattr(__import__("typing"), name, False)


def final(cls):
    """Decorator to make a class ``final``, i.e., it shouldn't be subclassed.

    It is the same as ``typing.final`` or an equivalent implementation
    depending on the python version and whether typing-extensions is
    installed.
    """
    setattr(cls, "__final__", True)
    return cls


def is_compatible_final(final) -> bool:
    @final
    class FinalClass:
        pass

    return getattr(FinalClass, "__final__", False)


stdlib_final = typing_extensions_import("final")
if stdlib_final and is_compatible_final(stdlib_final) and "SPHINX_BUILD" not in os.environ:
    final = stdlib_final  # noqa: F811


def import_typeshed_client():
    if typeshed_client_support:
        import typeshed_client

        return typeshed_client
    else:
        return __import__("argparse").Namespace(ImportedInfo=object, ModulePath=object, Resolver=object)


class UndefinedException(Exception):
    pass


def get_jsonschema_exceptions():
    from jsonschema.exceptions import ValidationError

    return (ValidationError,)


@contextmanager
def missing_package_raise(package, importer):
    try:
        yield None
    except ImportError as ex:
        raise ImportError(f"{package} package is required by {importer} :: {ex}") from ex


def import_jsonschema(importer):
    with missing_package_raise("jsonschema", importer):
        import jsonschema
    return jsonschema, jsonschema.Draft7Validator


def import_jsonnet(importer):
    with missing_package_raise("jsonnet", importer):
        import _jsonnet
    return _jsonnet


def import_requests(importer):
    with missing_package_raise("requests", importer):
        import requests
    return requests


def import_docstring_parser(importer):
    with missing_package_raise("docstring-parser", importer):
        import docstring_parser
    return docstring_parser


def import_fsspec(importer):
    with missing_package_raise("fsspec", importer):
        import fsspec
    return fsspec


def import_ruyaml(importer):
    with missing_package_raise("ruyaml", importer):
        import ruyaml
    return ruyaml


def import_reconplogger(importer):
    with missing_package_raise("reconplogger", importer):
        import reconplogger
    return reconplogger


def set_config_read_mode(
    urls_enabled: bool = False,
    fsspec_enabled: bool = False,
):
    """Enables/disables optional config read modes.

    Args:
        urls_enabled: Whether to read config files from URLs using requests package.
        fsspec_enabled: Whether to read config files from fsspec supported file systems.
    """
    imports = {
        "u": import_requests,
        "s": import_fsspec,
    }

    def update_mode(flag, enabled):
        global _config_read_mode
        if enabled:
            imports[flag]("set_config_read_mode")
            if flag not in _config_read_mode:
                _config_read_mode = _config_read_mode.replace("f", "f" + flag)
        else:
            _config_read_mode = _config_read_mode.replace(flag, "")

    update_mode("u", urls_enabled)
    update_mode("s", fsspec_enabled)


def get_config_read_mode() -> str:
    """Returns the current config reading mode."""
    return _config_read_mode


def set_docstring_parse_options(style=None, attribute_docstrings: Optional[bool] = None):
    """Sets options for docstring parsing.

    Args:
        style (docstring_parser.DocstringStyle): The docstring style to expect.
        attribute_docstrings: Whether to parse attribute docstrings (slower).
    """
    global _docstring_parse_options
    dp = import_docstring_parser("set_docstring_parse_options")
    if style is not None:
        if not isinstance(style, dp.DocstringStyle):
            raise ValueError(f"Expected style to be of type {dp.DocstringStyle}.")
        _docstring_parse_options["style"] = style
    if attribute_docstrings is not None:
        if not isinstance(attribute_docstrings, bool):
            raise ValueError("Expected attribute_docstrings to be boolean.")
        _docstring_parse_options["attribute_docstrings"] = attribute_docstrings


def get_docstring_parse_options():
    if _docstring_parse_options["style"] is None:
        dp = import_docstring_parser("get_docstring_parse_options")
        _docstring_parse_options["style"] = dp.DocstringStyle.AUTO
    return _docstring_parse_options


def parse_docstring(component, params=False, logger=None):
    dp = import_docstring_parser("parse_docstring")
    options = get_docstring_parse_options()
    try:
        if params and options["attribute_docstrings"]:
            return dp.parse_from_object(component, style=options["style"])
        else:
            return dp.parse(component.__doc__, style=options["style"])
    except (ValueError, dp.ParseError) as ex:
        if logger:
            logger.debug(f"Failed parsing docstring for {component}: {ex}")
    return None


def parse_docs(component, parent, logger):
    docs = {}
    if docstring_parser_support:
        doc_sources = [component]
        if inspect.isclass(parent) and component.__name__ == "__init__":
            doc_sources += [parent]
        for src in doc_sources:
            doc = parse_docstring(src, params=True, logger=logger)
            if doc:
                for param in doc.params:
                    docs[param.arg_name] = param.description
    return docs


def get_doc_short_description(function_or_class, method_name=None, logger=None):
    if docstring_parser_support:
        component = function_or_class
        if inspect.isclass(function_or_class):
            if not method_name:
                docstring = parse_docstring(function_or_class, params=False, logger=logger)
                if docstring and docstring.short_description:
                    return docstring.short_description
            component = getattr(function_or_class, method_name or "__init__")
        docstring = parse_docstring(component, params=False, logger=logger)
        if docstring:
            return docstring.short_description
    return None


def get_omegaconf_loader():
    """Returns a yaml loader function based on OmegaConf which supports variable interpolation."""
    import io

    from ._loaders_dumpers import yaml_load

    with missing_package_raise("omegaconf", "get_omegaconf_loader"):
        from omegaconf import OmegaConf

    def omegaconf_load(value):
        value_pyyaml = yaml_load(value)
        if isinstance(value_pyyaml, (str, int, float, bool)) or value_pyyaml is None:
            return value_pyyaml
        value_omegaconf = OmegaConf.to_object(OmegaConf.load(io.StringIO(value)))
        str_ref = {k: None for k in [value]}
        return value_pyyaml if value_omegaconf == str_ref else value_omegaconf

    return omegaconf_load


annotated_alias = typing_extensions_import("_AnnotatedAlias")


def is_annotated(typehint: type) -> bool:
    return annotated_alias and isinstance(typehint, annotated_alias)


def get_annotated_base_type(typehint: type) -> type:
    return typehint.__origin__  # type: ignore[attr-defined]


type_alias_type = typing_extensions_import("TypeAliasType")


def is_alias_type(typehint: type) -> bool:
    return type_alias_type and isinstance(typehint, type_alias_type)


def get_alias_target(typehint: type) -> bool:
    return typehint.__value__  # type: ignore[attr-defined]


def get_pydantic_support() -> int:
    support = "0"
    if find_spec("pydantic"):
        try:
            from importlib.metadata import version

            support = version("pydantic")
        except ImportError:
            import pydantic

            support = pydantic.version.VERSION

    return int(support.split(".", 1)[0])


pydantic_support = get_pydantic_support()


def get_pydantic_supports_field_init() -> bool:
    if find_spec("pydantic"):
        try:
            from importlib.metadata import version

            support = version("pydantic")
        except ImportError:
            import pydantic

            support = pydantic.version.VERSION
        major, minor = tuple(int(x) for x in support.split(".")[:2])
        return major > 2 or (major == 2 and minor >= 6)
    return False


pydantic_supports_field_init = get_pydantic_supports_field_init()


def is_pydantic_model(class_type) -> int:
    classes = inspect.getmro(class_type) if pydantic_support and inspect.isclass(class_type) else []
    for cls in classes:
        if getattr(cls, "__module__", "").startswith("pydantic") and getattr(cls, "__name__", "") == "BaseModel":
            import pydantic

            if issubclass(cls, pydantic.BaseModel):
                return pydantic_support
            elif pydantic_support > 1 and issubclass(cls, pydantic.v1.BaseModel):
                return 1
    return 0


def get_module(value):
    return getattr(type(value), "__module__", "").split(".", 1)[0]


def is_annotated_validator(typehint: type) -> bool:
    return (
        pydantic_support > 1
        and is_annotated(typehint)
        and any(get_module(m) in {"pydantic", "annotated_types"} for m in typehint.__metadata__)  # type: ignore[attr-defined]
    )


def validate_annotated(value, typehint: type):
    from pydantic import TypeAdapter

    return TypeAdapter(typehint).validate_python(value)



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class RecommendSizeAdjustmentsCLI(scfg.DataConfig):
    """
    Helper to recommend adjustments to network size parameters

    Example:

        MAX_STEPS=10000
        MAX_EPOCHS=720
        TRAIN_BATCHES_PER_EPOCH=2666
        ACCUMULATE_GRAD_BATCHES=32
        BATCH_SIZE=6
        TRAIN_ITEMS_PER_EPOCH=$(python -c "print($TRAIN_BATCHES_PER_EPOCH * $BATCH_SIZE)")

        python -m geowatch.cli.experimental.recommend_size_adjustments \
            --MAX_STEPS=$MAX_STEPS \
            --MAX_EPOCHS=$MAX_EPOCHS \
            --BATCH_SIZE=$BATCH_SIZE \
            --ACCUMULATE_GRAD_BATCHES=$ACCUMULATE_GRAD_BATCHES \
            --TRAIN_BATCHES_PER_EPOCH="$TRAIN_BATCHES_PER_EPOCH" \
            --TRAIN_ITEMS_PER_EPOCH="$TRAIN_ITEMS_PER_EPOCH"


    """
    MAX_STEPS               = scfg.Value(None, help='The number of optimizer steps to be taken')
    MAX_EPOCHS              = scfg.Value(None, help='The maximum number of train epochs')
    BATCH_SIZE              = scfg.Value(None, help='The physical batch size')
    ACCUMULATE_GRAD_BATCHES = scfg.Value(1, help='Accumulate gradients for this many batches for stepping the optimizer. I.e. the multiplier for effective batch size')
    TRAIN_ITEMS_PER_EPOCH   = scfg.Value(None, help='The number of items the training dataloader can produce in one epoch')
    TRAIN_BATCHES_PER_EPOCH = scfg.Value(None, help='The number of items the training dataloader can produce in one epoch')

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from geowatch.cli.experimental.recommend_size_adjustments import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = RecommendSizeAdjustmentsCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        from rich.markup import escape
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        rich.print('config = ' + escape(ub.urepr(config, nl=1, align=":")))

        import sympy

        assert config.BATCH_SIZE is not None

        if config.MAX_EPOCHS is None:
            ...

        if config.TRAIN_ITEMS_PER_EPOCH is None:
            config.TRAIN_ITEMS_PER_EPOCH = config.TRAIN_BATCHES_PER_EPOCH * config.BATCH_SIZE

        symbolic_names = 'TRAIN_ITEMS_PER_EPOCH, BATCH_SIZE, ACCUMULATE_GRAD_BATCHES, MAX_EPOCHS, MAX_STEPS'.split(', ')
        # symbolic_vars = sympy.symbols(symbolic_names, integer=True, positive=True)
        symbolic_vars = sympy.symbols(symbolic_names)
        TRAIN_ITEMS_PER_EPOCH, BATCH_SIZE, ACCUMULATE_GRAD_BATCHES, MAX_EPOCHS, MAX_STEPS = symbolic_vars

        # Build substitution dictionary for sympy
        subs = ub.dzip(symbolic_vars, ub.udict(config).take(symbolic_names))

        effective_batch_size = ACCUMULATE_GRAD_BATCHES * BATCH_SIZE
        steps_per_epoch = TRAIN_ITEMS_PER_EPOCH / effective_batch_size
        # This next line is more correct, but prevents the symbolic solver from
        # working. Can uncomment if we fixup the numeric solver to work better.
        # steps_per_epoch = sympy.floor(TRAIN_ITEMS_PER_EPOCH / effective_batch_size)
        total_steps = MAX_EPOCHS * steps_per_epoch
        total_steps.subs(subs)

        steps_per_epoch_ = steps_per_epoch.subs(subs).evalf()
        effective_batch_size_ = effective_batch_size.subs(subs).evalf()

        # The training progress iterator should show this number as the total number
        import math
        train_epoch_progbar_total_ = math.ceil((TRAIN_ITEMS_PER_EPOCH / BATCH_SIZE).subs(subs).evalf())

        print(f'steps_per_epoch_           = {steps_per_epoch_}')
        print(f'effective_batch_size_      = {effective_batch_size_}')
        print(f'train_epoch_progbar_total_ = {train_epoch_progbar_total_}')

        diff = MAX_STEPS - total_steps
        step_difference = diff.subs(subs)
        print(f'step_difference={step_difference.evalf()}')

        if step_difference == 0:
            rich.print(':white_check_mark: Parameters are perfectly balanced :balance_scale:')
        elif step_difference > 0:
            rich.print(':x: Not enough total steps to fill MAX_STEPS')
        else:
            rich.print(':x: MAX STEPS will stop training short')

        def numeric_solve(to_zero, k):
            from scipy.optimize import minimize

            def func(x):
                v = float(x[0])
                result = to_zero.subs({k: v}).evalf() ** 2
                return float(result)

            guess = config[str(k)]
            results = minimize(func, guess)
            return int(results.x[0])

        rich.print('[white]--- Possible Adjustments ---')
        for k, v in subs.items():
            tmp_subs = (ub.udict(subs) - {k})
            to_zero = diff.subs(tmp_subs)
            initial = config[str(k)]
            try:
                solutions = sympy.solve(to_zero, k)
                solutions = [s.evalf() for s in solutions]
                if len(solutions) == 0:
                    raise Exception
                suggestion = solutions
                method = 'symbolic'
            except Exception:
                numeric_solution = numeric_solve(to_zero, k)
                suggestion = numeric_solution
                method = 'numeric'
            rich.print(f' * {k}: {initial} -> {suggestion} ({method})')


__cli__ = RecommendSizeAdjustmentsCLI
main = __cli__.main

if __name__ == '__main__':
    """

    CommandLine:
        python -m geowatch.cli.experimental.recommend_size_adjustments --help
    """
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple wrapper of hierarchical dictionary for hparams."""

import ast
import collections
import copy
from typing import Any, Dict, Text, List, Callable, Optional

import tensorflow as tf
import yaml

REQUIRED = '__required__'


def eval_str_fn(val: str) -> Any:
    if '|' in val:
        return [eval_str_fn(v) for v in val.split('|')]
    if val in {'true', 'false'}:
        return val == 'true'
    try:
        return ast.literal_eval(val)
    except (ValueError, SyntaxError):
        return val


# pylint: disable=protected-access
class Config(dict):
    """A config utility class."""

    def __init__(self, *args, **kwargs):
        super().__init__()
        input_config_dict = dict(*args, **kwargs)
        self.update(input_config_dict)

    def __len__(self):
        return len(self.__dict__)

    def __setattr__(self, k: str, v: Any) -> None:
        if isinstance(v, dict) and not isinstance(v, Config):
            self.__dict__[k] = Config(v)
        else:
            self.__dict__[k] = copy.deepcopy(v)

    def __getattr__(self, k: str) -> Any:
        return self.__dict__[k]

    def __setitem__(self, k: str, v: Any) -> None:
        self.__setattr__(k, v)

    def __getitem__(self, k: str) -> Any:
        return self.__dict__[k]

    def __contains__(self, k: str) -> bool:
        return self.__dict__.__contains__(k)

    def __iter__(self):
        for key in self.__dict__:
            yield key

    def items(self):
        for key, value in self.__dict__.items():
            yield key, value

    def replace(self, **kwargs) -> 'Config':
        """Deep copies and replaces some values."""
        cfg = copy.deepcopy(self)
        cfg.update(dict(**kwargs))
        return cfg

    def __repr__(self):
        return repr(self.as_dict())

    def __getstate__(self):
        return self.__dict__

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo: Any) -> 'Config':
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            result[k] = v
        return result

    def __str__(self):
        try:
            return yaml.dump(self.as_dict(), indent=4)
        except TypeError:
            return str(self.as_dict())

    def _update(self, config_dict, allow_new_keys=True, skip_new_keys=False):
        """Recursively updates internal members."""
        if not config_dict:
            return

        for k, v in config_dict.items():
            if k not in self.__dict__:
                if allow_new_keys:
                    self.__setattr__(k, v)
                elif skip_new_keys:
                    pass
                else:
                    raise KeyError('Key `{}` does not exist for overriding. '.format(k))
            else:
                if isinstance(self.__dict__[k], Config) and isinstance(v, dict):
                    self.__dict__[k]._update(v, allow_new_keys)
                elif isinstance(self.__dict__[k], Config) and isinstance(v, Config):
                    self.__dict__[k]._update(v.as_dict(), allow_new_keys)
                else:
                    self.__setattr__(k, v)

    def get(self, k: str, default_value: Any = None) -> Any:
        return self.__dict__.get(k, default_value)

    def update(self, config_dict: Optional[Dict[Any, Any]] = None) -> None:
        """Updates members while allowing new keys."""
        if config_dict:
            self._update(config_dict, allow_new_keys=True)

    def update_dict(self, **kwargs):
        """Updates members while allowing new keys."""
        self._update(kwargs, allow_new_keys=True)

    def keys(self) -> Any:
        return self.__dict__.keys()

    def override_dict(self, skip_new_keys: bool = True, **kwargs) -> None:
        """Overrides members and skips new keys."""
        self._update(kwargs, allow_new_keys=False, skip_new_keys=skip_new_keys)

    def override(self, config_dict_or_str: ...,
                 allow_new_keys: bool = False) -> None:
        """Updates members while disallowing new keys."""
        if not config_dict_or_str:
            return
        if isinstance(config_dict_or_str, str):
            if '=' in config_dict_or_str:
                config_dict = self.parse_from_str(config_dict_or_str)
            elif config_dict_or_str.endswith('.yaml'):
                config_dict = self.parse_from_yaml(config_dict_or_str)
            else:
                raise ValueError(
                    'Invalid string {}, must end with .yaml or contains "=".'.format(
                        config_dict_or_str))
        elif isinstance(config_dict_or_str, dict):
            config_dict = config_dict_or_str
        else:
            raise ValueError('Unknown value type: {}'.format(config_dict_or_str))

        self._update(config_dict, allow_new_keys)

    def parse_from_yaml(self, yaml_file_path: Text) -> Dict[Any, Any]:
        """Parses a yaml file and returns a dictionary."""
        with tf.io.gfile.GFile(yaml_file_path, 'r') as f:
            config_dict = yaml.load(f, Loader=yaml.FullLoader)
            return config_dict

    def save_to_yaml(self, yaml_file_path: str) -> None:
        """Writes a dictionary into a yaml file."""
        with tf.io.gfile.GFile(yaml_file_path, 'w') as f:
            yaml.dump(self.as_dict(), f, default_flow_style=False)

    def parse_from_str(self, config_str: Text) -> Dict[Any, Any]:
        """Parses a string like 'x.y=1,x.z=2' to nested dict {x: {y: 1, z: 2}}."""
        if not config_str:
            return {}
        config_dict = {}
        try:
            for kv_pair in config_str.split(','):
                # We skip the empty string here.
                if not kv_pair:
                    continue
                key_str, value_str = kv_pair.split('=')
                key_str = key_str.strip()

                def add_kv_recursive(k, v):
                    """Recursively parses x.y.z=tt to {x: {y: {z: tt}}}."""
                    if '.' not in k:
                        return {k: eval_str_fn(v)}
                    pos = k.index('.')
                    return {k[:pos]: add_kv_recursive(k[pos + 1:], v)}  # pylint: disable=cell-var-from-loop

                def merge_dict_recursive(target, src):
                    """Recursively merges two nested dictionary."""
                    for k in src.keys():
                        if ((k in target and isinstance(target[k], dict) and
                             isinstance(src[k], collections.abc.Mapping))):
                            merge_dict_recursive(target[k], src[k])  # pylint: disable=cell-var-from-loop
                        else:
                            target[k] = src[k]

                merge_dict_recursive(config_dict, add_kv_recursive(key_str, value_str))
            return config_dict
        except ValueError:
            raise ValueError(f'Invalid config_str: {config_str}') from None

    def as_dict(self) -> Dict[Any, Any]:
        """Returns a dict representation."""
        config_dict = {}
        for k, v in self.__dict__.items():
            if isinstance(v, Config):
                config_dict[k] = v.as_dict()
            elif isinstance(v, (list, tuple)):
                config_dict[k] = [
                    i.as_dict() if isinstance(i, Config) else copy.deepcopy(i)
                    for i in v
                ]
            else:
                config_dict[k] = copy.deepcopy(v)
        return config_dict

    def validate(self):
        required_keys = []
        for k, v in self.__dict__.items():
            if v == REQUIRED:
                required_keys.append(k)
        if required_keys:
            raise ValueError(f'Values are required for keys: {required_keys}')


class RegistryFactor():
    """A template for registry factory."""

    def __init__(self, prefix: str) -> None:
        self.registry_map = {}
        self.prefix = prefix

    def register(self, name: Optional[str] = None) -> Callable[[Any], Any]:
        """Registers a function, mainly for config here."""

        def decorator(cls):
            key = self.prefix + (name or cls.__name__.lower())
            if key in self.registry_map:
                raise ValueError(f'{key} is already registered')
            self.registry_map[key] = cls
            return cls

        return decorator

    def lookup(self, name: str) -> Config:
        """Looks up a class based on class name."""
        key = self.prefix + name.lower()
        if key not in self.registry_map:
            raise KeyError(f'{key} is not in {self.registry_map.keys()}')
        return self.registry_map[key]

    def keys(self, prefix: str = '') -> List[Config]:
        return [
            k[len(prefix):]
            for k in self.registry_map.keys()
            if k.startswith(prefix)
        ]


def create_config_from_dict(config_dict: Dict[str, Any],
                            required_keys: List[str],
                            optional_keys: Dict[str, Any]) -> Config:
    """Creates hparam config from dictionary."""
    config = Config()
    for key in required_keys:
        if key not in config_dict:
            raise ValueError('Required key %s missed in config dict.' % key)
        config[key] = config_dict[key]
    for key in optional_keys:
        if key not in config_dict:
            config[key] = optional_keys[key]
        else:
            config[key] = config_dict[key]
    return config



#!/usr/bin/env python3
r"""
Given a set of site summaries, clusters them into groups, ideally with small
overlap. Writes new regions to a specified directory using the hash of the
contained sites as a subregion identifier.

Limitations:
    - The clustering algorithm is overly simple

Example:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m geowatch.cli.cluster_sites \
            --src "$DVC_DATA_DPATH/annotations/drop6/region_models/KR_R002.geojson" \
            --dst_dpath $DVC_DATA_DPATH/ValiRegionSmall/geojson \
            --draw_clusters True

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m geowatch.cli.coco_align \
        --src $DVC_DATA_DPATH/Drop6/combo_imganns-KR_R002_L.kwcoco.json \
        --dst $DVC_DATA_DPATH/ValiRegionSmall/small_KR_R002_odarcigm.kwcoco.zip \
        --regions $DVC_DATA_DPATH/ValiRegionSmall/geojson/SUB_KR_R002_n007_odarcigm.geojson \
        --minimum_size="128x128@10GSD" \
        --context_factor=1 \
        --geo_preprop=auto \
        --force_nodata=-9999 \
        --site_summary=False \
        --target_gsd=5 \
        --aux_workers=8 \
        --workers=8

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m geowatch.cli.cluster_sites \
            --src "$DVC_DATA_DPATH/annotations/drop6/region_models/NZ_R001.geojson" \
            --dst_dpath $DVC_DATA_DPATH/ValiRegionSmall/geojson/NZ_R001 \
            --draw_clusters True

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m geowatch.cli.coco_align \
        --src $DVC_DATA_DPATH/Drop6/combo_imganns-NZ_R001_L.kwcoco.json \
        --dst $DVC_DATA_DPATH/ValiRegionSmall/small_NZ_R001_swnykmah.kwcoco.zip \
        --regions $DVC_DATA_DPATH/ValiRegionSmall/geojson/NZ_R001/SUB_NZ_R001_n031_swnykmah.geojson \
        --minimum_size="128x128@10GSD" \
        --context_factor=1 \
        --geo_preprop=auto \
        --force_nodata=-9999 \
        --site_summary=False \
        --target_gsd=5 \
        --aux_workers=8 \
        --workers=8
"""
import scriptconfig as scfg
import ubelt as ub


class ClusterSiteConfig(scfg.DataConfig):
    """
    Creates a new region file that groups nearby sites.
    """
    src = scfg.Value(None, help='input region files with site summaries', alias=['regions'])

    dst_dpath = scfg.Value(None, help='output path to store the resulting region files')

    dst_region_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        geojson output path for site summaries. Only works if there is a single
        input region, otherwise an error is raised
        '''))

    io_workers = scfg.Value(10, help='number of io workers')

    draw_clusters = scfg.Value(False, isflag=True, help='if True draw the clusters in the specified dst_dpath')

    crop_time = scfg.Value(True, isflag=True, help=ub.paragraph(
        '''
        if True also crops temporal extent to the sites, otherwise uses the region extent
        '''))

    minimum_size = scfg.Value('128x128@2GSD', help='minimum size of a cluster box')
    maximum_size = scfg.Value('1024x1024@2GSD', help='minimum size of a cluster box')

    context_factor = scfg.Value(1.5, help='extra padding around each site')

    ignore_status = scfg.Value("['system_rejected']", type=str, help=ub.paragraph(
        '''
        A YAML list of status values that should be ignored by the clustering
        algorithm.  Defaults to ["system_rejected"].
        '''))


def main(cmdline=1, **kwargs):
    """
    CommandLine:
        xdoctest -m geowatch.cli.cluster_sites main:0
        xdoctest -m geowatch.cli.cluster_sites main:2

    Example:
        >>> from geowatch.cli.cluster_sites import *  # NOQA
        >>> from geowatch.cli import cluster_sites
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('geowatch', 'doctests', 'cluster_sites1').ensuredir()
        >>> src_dpath = (dpath / 'src').ensuredir()
        >>> dst_dpath = (dpath / 'dst')
        >>> dst_region_fpath = dst_dpath / 'cluster.geojson'
        >>> from geowatch.geoannots import geomodels
        >>> region = geomodels.RegionModel.random(num_sites=10)
        >>> src_fpath = src_dpath / 'demo_region.geojson'
        >>> src_fpath.write_text(region.dumps())
        >>> kwargs = {
        >>>     'src': src_fpath,
        >>>     'dst_dpath': dst_dpath,
        >>>     'dst_region_fpath': dst_region_fpath,
        >>>     'io_workers': 0,
        >>>     'draw_clusters': 0,
        >>>     'crop_time': True,
        >>> }
        >>> cmdline = 0
        >>> cluster_sites.main(cmdline=cmdline, **kwargs)

    Example:
        >>> from geowatch.cli.cluster_sites import *  # NOQA
        >>> from geowatch.cli import cluster_sites
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('geowatch', 'doctests', 'cluster_sites2').ensuredir()
        >>> src_dpath = (dpath / 'src').ensuredir()
        >>> dst_dpath = (dpath / 'dst')
        >>> from geowatch.geoannots import geomodels
        >>> region = geomodels.RegionModel.random(num_sites=10)
        >>> src_fpath = src_dpath / 'demo_region.geojson'
        >>> src_fpath.write_text(region.dumps())
        >>> dst_region_fpath = dst_dpath / 'cluster.geojson'
        >>> kwargs = {
        >>>     'src': src_fpath,
        >>>     'dst_dpath': dst_dpath,
        >>>     'dst_region_fpath': dst_region_fpath,
        >>>     'io_workers': 0,
        >>>     'draw_clusters': 1,
        >>>     'crop_time': True,
        >>> }
        >>> cmdline = 0
        >>> cluster_sites.main(cmdline=cmdline, **kwargs)

    Example:
        >>> # Test empty case
        >>> from geowatch.cli.cluster_sites import *  # NOQA
        >>> from geowatch.cli import cluster_sites
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('geowatch', 'doctests', 'cluster_sites3').ensuredir()
        >>> src_dpath = (dpath / 'src').ensuredir()
        >>> dst_dpath = (dpath / 'dst')
        >>> from geowatch.geoannots import geomodels
        >>> region = geomodels.RegionModel.random(num_sites=0)
        >>> src_fpath = src_dpath / 'demo_region.geojson'
        >>> src_fpath.write_text(region.dumps())
        >>> dst_region_fpath = dst_dpath / 'cluster.geojson'
        >>> kwargs = {
        >>>     'src': src_fpath,
        >>>     'dst_dpath': dst_dpath,
        >>>     'dst_region_fpath': dst_region_fpath,
        >>>     'io_workers': 0,
        >>>     'draw_clusters': 1,
        >>>     'crop_time': True,
        >>> }
        >>> cmdline = 0
        >>> cluster_sites.main(cmdline=cmdline, **kwargs)

    Ignore:
        import xdev
        xdev.profile_now(cluster_sites.main)(cmdline=cmdline, **kwargs)

    Ignore:
        import sys, ubelt
        sys.path.append(ubelt.expandpath('~/code/watch'))
        from geowatch.cli.cluster_sites import *  # NOQA
        import geowatch
        data_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        src = data_dpath / 'annotations/drop6/region_models/KR_R002.geojson'
        dst_dpath = data_dpath / 'ValiRegionSmall/geojson'
        kwargs = dict(src=src, dst_dpath=dst_dpath, draw_clusters=True)
        main(**kwargs)
    """
    config = ClusterSiteConfig.cli(cmdline=cmdline, data=kwargs)
    import rich
    from rich.markup import escape
    rich.print('config = {}'.format(escape(ub.urepr(config, nl=1))))

    from kwutil import util_yaml
    from geowatch.geoannots import geomodels
    from geowatch.utils import process_context
    from geowatch.utils import util_resolution

    config.ignore_status = util_yaml.Yaml.coerce(config.ignore_status)
    print(f'config.ignore_status = {ub.urepr(config.ignore_status, nl=1)}')

    # import pandas as pd
    if config.dst_dpath is None:
        raise ValueError('Destination path is required')

    dst_dpath = config.dst_dpath = ub.Path(config.dst_dpath)
    config.dst_dpath.ensuredir()
    rich.print(f'Will write to: [link={dst_dpath}]{dst_dpath}[/link]')

    proc_context = process_context.ProcessContext(
        name='geowatch.cli.cluster_sites', type='process',
        config=process_context.jsonify_config(config),
        track_emissions=False,
    )

    input_region_models = list(geomodels.RegionModel.coerce_multiple(
        config.src, workers=config.io_workers))

    if config.dst_region_fpath is not None:
        if len(input_region_models) != 1:
            raise ValueError(
                'we assume only 1 input region when output fpath given')
        # Fixme: process context is weird here when we allow for multiple
        # region inputs
        proc_context.start()

    scale = config.context_factor
    minimum_size = util_resolution.ResolvedWindow.coerce(config.minimum_size)
    maximum_size = util_resolution.ResolvedWindow.coerce(config.maximum_size)

    # Convert to meters
    meter = util_resolution.ResolvedUnit(1, 'GSD')
    min_box_dim = minimum_size.at_resolution(meter).window[0]
    max_box_dim = maximum_size.at_resolution(meter).window[0]

    print(f'Looping over {len(input_region_models)} region')

    # all_final_site_summaries = []

    # Note: this should only ever be a single item in this loop.  outputs will
    # clobber each other. The code is setup to allow flexability to draw
    # multiple region clusters at once, but we might want to disable this.
    for input_region_model in input_region_models:
        cluster_single_region_sites(input_region_model, scale, min_box_dim,
                                    max_box_dim, proc_context, config)


def cluster_single_region_sites(input_region_model, scale, min_box_dim, max_box_dim, proc_context, config):
    import kwimage
    from kwgis.utils import util_gis
    from geowatch.utils import util_kwimage
    input_region_model.fixup()

    if 1:
        try:
            input_region_model.validate(strict=0)
        except Exception:
            input_region_model.fixup()
            input_region_model.validate(strict=0)

    # Create the set of input boxes to the clustering algorithm.
    # Filter out any that have an ignorable status, and convert to UTM
    # coordinates.
    region_header = input_region_model.pandas_region()
    region_sites = input_region_model.pandas_summaries()
    assert len(region_header) == 1
    if config.ignore_status:
        # Remove any sites that have a status marked as ignored.
        if 'status' in region_sites.columns:
            keep_flags = region_sites['status'].apply(lambda s: s not in config.ignore_status)
            # For some reason this returns a dataframe instead of a
            # geodataframe if all flags are false Using an index with loc seems
            # to work around this.
            keep_indexes = keep_flags.index[keep_flags.index]
            region_sites = region_sites.loc[keep_indexes]

    region_sites_utm = util_gis.project_gdf_to_local_utm(region_sites, mode=1, tolerance=None)
    polygons = kwimage.PolygonList([kwimage.Polygon.from_shapely(s)
                                    for s in region_sites_utm.geometry])

    # Run the clustering algorithm.
    # TODO: would be good to do this in 3D with temporal extent too.
    keep_bbs, overlap_idxs = util_kwimage.find_low_overlap_covering_boxes(
        polygons, scale, min_box_dim, max_box_dim, max_iters=100)

    # Sort bbs so the largest spatial ones are first.
    keep_bbs = keep_bbs[keep_bbs.area.ravel().argsort()[::-1]]

    # Convert the UTM bounding boxes into proper region models.
    clustered_region, subregions = build_clustered_models(
        input_region_model, region_header, region_sites, region_sites_utm,
        keep_bbs, config)

    # Write output to disk
    # Awkward backwards compat, dump individal cluster groups
    for subregion in subregions:
        fpath = config.dst_dpath / (subregion.region_id + '.geojson')
        fpath.write_text(subregion.dumps())

    if config.dst_region_fpath is not None:
        obj = proc_context.stop()
        clustered_region.header['properties']['cache']['proc_context'] = obj
        region_fpath = ub.Path(config.dst_region_fpath)
        region_fpath.write_text(clustered_region.dumps())

    if config.draw_clusters:
        _draw_clusters(input_region_model, region_sites_utm, polygons,
                       clustered_region, keep_bbs, config)


def _draw_clusters(input_region_model, region_sites_utm, polygons,
                   clustered_region, keep_bbs, config):
    from geowatch import heuristics
    from geowatch.utils import util_kwplot
    import kwimage
    import kwplot
    import rich

    if 'status' in region_sites_utm.columns:
        color_list = []
        status_list = region_sites_utm['status']
        for status in status_list:
            info = heuristics.IARPA_STATUS_TO_INFO.get(status, {})
            color = kwimage.Color.coerce(info.get('color', 'pink')).as255()
            color_list.append(color)
    else:
        color_list = ['blue'] * len(polygons)

    plt = kwplot.autoplt()
    kwplot.figure(fnum=1, doclf=1)
    # polygons.draw(color='pink')
    for poly, color in zip(polygons, color_list):
        edgecolor = kwimage.Color.coerce(color).adjust(saturate=-.1, lighten=.1)
        poly.draw(color=color, edgecolor=edgecolor, linewidth=1, alpha=0.7)
    # candidate_bbs.draw(color='blue', setlim=1)
    subregion_suffix_list = [
        ss['properties']['cache']['cluster_suffix']
        for ss in clustered_region.site_summaries()
    ]
    if len(keep_bbs):
        keep_bbs.draw(color='orange', setlim=1, labels=subregion_suffix_list)
    plt.gca().set_title('find_low_overlap_covering_boxes')
    fig = plt.gcf()
    viz_dpath = (config.dst_dpath / '_viz_clusters').ensuredir()
    rich.print(f'Viz dpath: [link={viz_dpath}]{viz_dpath}[/link]')

    finalizer = util_kwplot.FigureFinalizer(
        size_inches=(16, 16),
        dpath=viz_dpath,
    )
    finalizer(fig, 'clusters_' + input_region_model.region_id + '.png')


def build_clustered_models(input_region_model, region_header, region_sites,
                           region_sites_utm, keep_bbs, config):
    """
    Given the clustering output, construct new region models.

    Returns:
        Tuple[RegionModel, List[RegionModle]] :

        1. A new "clustered region" region model with the original region
           bounds and the clusters as site summaries

        2. A set of region models for each cluster, with the original site
           summaries that were assigned to that cluster

    """
    from geowatch.geoannots import geomodels
    import geopandas as gpd

    region_id = input_region_model.region_id

    utm_crs = region_sites_utm.crs
    crs84 = region_header.crs

    # Start an empty region model with the original bounds
    # we will add clusters as site summaries to this data structure.
    new_region_header = geomodels.RegionHeader(
        properties={
            "type": 'region',
            "region_id": input_region_model.region_id,
            "version": '2.4.3',
            "mgrs": None,
            "start_date": input_region_model.start_date.date().isoformat(),
            "end_date": input_region_model.end_date.date().isoformat(),
            "originator": "kit-cluster",
            "model_content": "annotation",
            "comments": '',
            "cache": {}
        },
        geometry=input_region_model.geometry,
    )
    new_region_header.ensure_isodates()
    new_region_header.infer_mgrs()
    clustered_region = geomodels.RegionModel(features=[new_region_header])

    print(f'Looping over {len(keep_bbs)} clusters')
    subregions = []
    for bb_idx, utm_box in enumerate(keep_bbs.to_shapely()):

        ISECT_BOXES = 0
        if ISECT_BOXES:
            region_utm = region_header.to_crs(utm_crs)
            region_bounds = region_utm.iloc[0]['geometry']
            final_geom_utm = region_bounds.intersection(utm_box)
        else:
            final_geom_utm = utm_box

        # Convert the UTM cluster into CRS84
        final_geom_df_crs84 = gpd.GeoDataFrame(
            {'geometry': [final_geom_utm]}, crs=utm_crs).to_crs(crs84)
        final_geom_crs84 = final_geom_df_crs84.iloc[0]['geometry']
        cluster_geometry = final_geom_crs84
        if (cluster_geometry is None or
             cluster_geometry.is_empty or
             not cluster_geometry.is_valid):
            raise AssertionError

        # Assign original sites to this cluster.
        is_contained = region_sites_utm.intersects(final_geom_utm)
        contained_sites = region_sites[is_contained]

        cluster_suffix = f'{bb_idx:03d}'
        cluster_id = f'{region_id}_CLUSTER_{cluster_suffix}'
        start_dates = contained_sites['start_date'].dropna()
        end_dates = contained_sites['end_date'].dropna()

        site_summaries = list(geomodels.SiteSummary.from_geopandas_frame(contained_sites))
        if 1:
            for s in site_summaries:
                s.fixup()
                s.validate(strict=0)

        # The start / end date of the cluster is determined by the original
        # sites that intersect it.
        if len(start_dates) and config.crop_time:
            start_date = start_dates.min()
        else:
            start_date = input_region_model.start_date
        if len(end_dates) and config.crop_time:
            end_date = end_dates.max()
        else:
            end_date = input_region_model.end_date

        # Add this cluster as a site summary in the new region
        cluster_summary = geomodels.SiteSummary(
            properties={
                "type": 'site_summary',
                "status": "system_confirmed",
                "site_id": cluster_id,
                "version": '2.4.3',
                "mgrs": None,
                "start_date": start_date,
                "end_date": end_date,
                "originator": "kit-cluster",
                "model_content": "annotation",
                "cache": {"cluster_suffix": cluster_suffix},
            },
            geometry=cluster_geometry,
        )
        cluster_summary.ensure_isodates()
        cluster_summary.infer_mgrs()
        clustered_region.add_site_summary(cluster_summary)

        # Create a new region model where the cluster is the region boundary
        # that contatain the original site that were assigned to it.
        subregion_header = geomodels.RegionHeader(
            properties={
                "type": 'region',
                "region_id": cluster_id,
                "version": '2.4.3',
                "mgrs": None,
                "start_date": start_date,
                "end_date": end_date,
                "originator": "kit-cluster",
                "model_content": "annotation",
                "comments": '',
            },
            geometry=cluster_geometry,
        )
        assert subregion_header.geometry is not None
        subregion_header.ensure_isodates()
        subregion_header.infer_mgrs()
        subregion = geomodels.RegionModel(features=[
            subregion_header
        ] + site_summaries)
        try:
            subregion.validate(strict=False)
        except Exception:
            subregion.fixup()
            subregion.validate(strict=False)
        subregions.append(subregion)

    return clustered_region, subregions


__cli__ = ClusterSiteConfig
__cli__.main = main


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch/cli/cluster_sites.py
    """
    main(cmdline=True)





def test_reproject_annotations_positive_pending():
    from geowatch.geoannots.geomodels import RegionModel, SiteModel
    import kwarray

    start_time = '2012-03-04'
    end_time = '2025-06-07'

    region = RegionModel.random(
        with_sites=False,
        rng=1089283,
        start_time=start_time,
        end_time=end_time,
        num_sites=0,
    )

    sites = []
    rng = kwarray.ensure_rng(60810907460)

    N = 10

    # Create a positive_pending annotation with a null end date
    for _ in range(N):
        site = SiteModel.random(
            region=region, num_observations=2, rng=rng,
            start_site_index=len(sites),
        )
        obs1, obs2 = list(site.observations())
        obs1['properties']['current_phase'] = None
        obs2['properties']['current_phase'] = None
        obs2['properties']['observation_date'] = None
        site.header['properties']['status'] = 'positive_pending'
        site.header['properties']['end_date'] = None
        sites.append(site)

    # Create a positive_pending annotation with a null start date
    for _ in range(N):
        site = SiteModel.random(
            region=region, num_observations=2, rng=rng,
            start_site_index=len(sites),
        )
        obs1, obs2 = list(site.observations())
        obs1['properties']['current_phase'] = None
        obs2['properties']['current_phase'] = None
        obs1['properties']['observation_date'] = None
        site.header['properties']['status'] = 'positive_pending'
        site.header['properties']['start_date'] = None
        sites.append(site)

    # Create a positive_pending annotation with null start and end dates
    for _ in range(N):
        site = SiteModel.random(
            region=region, num_observations=2, rng=rng,
            start_site_index=len(sites),
        )
        obs1, obs2 = list(site.observations())
        obs1['properties']['current_phase'] = None
        obs2['properties']['current_phase'] = None
        obs2['properties']['observation_date'] = None
        obs1['properties']['observation_date'] = None
        site.header['properties']['status'] = 'positive_pending'
        site.header['properties']['end_date'] = None
        site.header['properties']['start_date'] = None
        sites.append(site)

    # Create a positive_pending annotation with both dates
    for _ in range(N):
        site = SiteModel.random(
            region=region, num_observations=2, rng=rng,
            start_site_index=len(sites),
        )
        obs1, obs2 = list(site.observations())
        obs1['properties']['current_phase'] = None
        obs2['properties']['current_phase'] = None
        site.header['properties']['status'] = 'positive_pending'
        sites.append(site)

    summaries = [s.as_summary() for s in sites]
    region['features'].extend(summaries)

    summaries_df = region.pandas_summaries()
    print(summaries_df[['site_id', 'start_date', 'end_date']])

    # Create a dummy kwcoco file that is georegistered to the random region
    import geowatch
    dset = geowatch.coerce_kwcoco(
        'geowatch-msi', num_videos=1,
        geodata={'region_geom': region.geometry},
        dates={'start_time': start_time, 'end_time': end_time}
    )
    dset.clear_annotations()

    dset.images().lookup('date_captured')

    # print(f'dset.fpath={dset.fpath}')
    # gdalinfo /home/joncrall/.cache/geowatch/demo_kwcoco_bundles/watch_vidshapes_a86db771/_assets/auxiliary/aux_B10_B11/img_00010.tif
    # gdalinfo /home/joncrall/.cache/geowatch/demo_kwcoco_bundles/watch_vidshapes_5b7fdde0/_assets/auxiliary/aux_B10_B11/img_00010.tif

    from geowatch.cli import reproject_annotations
    import geowatch
    import ubelt as ub
    viz_dpath = ub.Path.appdir('geowatch/tests/test_reproject/viz').ensuredir()
    cmdline = False
    kwargs = {
        'src': dset,
        'dst': 'return',
        'viz_dpath': viz_dpath,
        'workers': 4,
        'io_workers': 8,
        'site_models': sites,
        'region_models': region,
    }
    new_dset = reproject_annotations.main(cmdline=cmdline, **kwargs)
    print(f'new_dset={new_dset}')

    # TODO: check assertions, but the viz looks ok



import fnmatch
import os.path
import random
from collections import defaultdict

import shapely.geometry
import shapely.wkt
import multiprocess
import skimage.io
import numpy as np

from tqdm import tqdm

import torch
import torch.utils.data
import torchvision  # NOQA

from lydorn_utils import print_utils
from lydorn_utils import python_utils

from torch_lydorn.torchvision.datasets import utils


class xView2Dataset(torch.utils.data.Dataset):
    """
    xView2 xBD dataset: https://xview2.org/
    """

    def __init__(self, root: str, fold: str = "train", pre_process: bool = True,
                 patch_size: int = None,
                 pre_transform=None, transform=None, small: bool = False, pool_size: int = 1, raw_dirname: str = "raw",
                 processed_dirname: str = "processed"):
        """

        @param root:
        @param fold:
        @param pre_process: If True, the dataset will be pre-processed first, saving training patches on disk. If False, data will be serve on-the-fly without any patching.
        @param patch_size:
        @param pre_transform:
        @param transform:
        @param small: If True, use a small subset of the dataset (for testing)
        @param pool_size:
        @param processed_dirname:
        """
        self.root = root
        self.fold = fold
        self.pre_process = pre_process
        self.patch_size = patch_size
        self.pre_transform = pre_transform
        self.transform = transform
        self.small = small
        if self.small:
            print_utils.print_info("INFO: Using small version of the xView2 xBD dataset.")
        self.pool_size = pool_size
        self.raw_dirname = raw_dirname

        if self.pre_process:
            # Setup of pre-process
            self.processed_dirpath = os.path.join(self.root, processed_dirname, self.fold)
            stats_filepath = os.path.join(self.processed_dirpath, "stats-small.pt" if self.small else "stats.pt")
            processed_relative_paths_filepath = os.path.join(self.processed_dirpath,
                                                             "processed_paths-small.json" if self.small else "processed_paths.json")

            # Check if dataset has finished pre-processing by checking processed_relative_paths_filepath:
            if os.path.exists(processed_relative_paths_filepath):
                # Process done, load stats and processed_relative_paths
                self.stats = torch.load(stats_filepath)
                self.processed_relative_paths = python_utils.load_json(processed_relative_paths_filepath)
            else:
                # Pre-process not finished, launch it:
                tile_info_list = self.get_tile_info_list()
                self.stats = self.process(tile_info_list)
                # Save stats
                torch.save(self.stats, stats_filepath)
                # Save processed_relative_paths
                self.processed_relative_paths = [tile_info["processed_relative_filepath"] for tile_info in tile_info_list]
                python_utils.save_json(processed_relative_paths_filepath, self.processed_relative_paths)
        else:
            # Setup data sample list
            self.tile_info_list = self.get_tile_info_list()

    def get_tile_info_list(self):
        tile_info_list = []
        fold_dirpath = os.path.join(self.root, self.raw_dirname, self.fold)
        images_dirpath = os.path.join(fold_dirpath, "images")
        image_filenames = fnmatch.filter(os.listdir(images_dirpath), "*_pre_disaster.png")
        image_filenames = sorted(image_filenames)
        disaster_samples_dict = defaultdict(int)
        for image_filename in image_filenames:
            name_split = image_filename.split("_")
            disaster = name_split[0]
            if self.small:
                disaster_samples_dict[disaster] += 1
                if 10 < disaster_samples_dict[disaster]:
                    continue  # Skip this sample as there is already enough for the small dataset
            number = int(name_split[1])
            tile_info = {
                "name": f"{disaster}_{number:08d}",
                "disaster": disaster,
                "number": number,
                "image_filepath": os.path.join(fold_dirpath, "images", f"{disaster}_{number:08d}_pre_disaster.png"),
                "label_filepath": os.path.join(fold_dirpath, "labels", f"{disaster}_{number:08d}_pre_disaster.json"),
                "processed_relative_filepath": os.path.join(disaster, f"{number:08d}.pt")
            }
            tile_info_list.append(tile_info)
        return tile_info_list

    def process(self, tile_info_list):
        # os.makedirs(os.path.join(self.root, self.processed_dirname), exist_ok=True)
        with multiprocess.Pool(self.pool_size) as p:
            list_of_stats = list(
                tqdm(p.imap(self._process_one, tile_info_list), total=len(tile_info_list), desc="Process"))

        # Aggregate stats
        mean_per_disaster = defaultdict(list)
        var_per_disaster = defaultdict(list)
        class_freq = []
        for stats in list_of_stats:
            mean_per_disaster[stats["disaster"]].append(stats["mean"])
            var_per_disaster[stats["disaster"]].append(stats["var"])
            class_freq.append(stats["class_freq"])
        stats = {
            "mean": {},
            "std": {},
            "class_freq": None
        }
        for disaster in mean_per_disaster.keys():
            stats["mean"][disaster] = np.mean(np.stack(mean_per_disaster[disaster], axis=0), axis=0)
            stats["std"][disaster] = np.sqrt(np.mean(np.stack(var_per_disaster[disaster], axis=0), axis=0))
        stats["class_freq"] = np.mean(np.stack(class_freq, axis=0), axis=0)

        return stats

    def load_raw_data(self, tile_info):
        # Image:
        tile_info["image"] = skimage.io.imread(tile_info["image_filepath"])
        assert len(tile_info["image"].shape) == 3 and tile_info["image"].shape[
            2] == 3, f"image should have shape (H, W, 3), not {tile_info['image'].shape}..."

        # Annotations:
        label_json = python_utils.load_json(tile_info["label_filepath"])
        features_xy = label_json["features"]["xy"]
        tile_info["gt_polygons"] = [shapely.wkt.loads(obj["wkt"]) for obj in features_xy]

        return tile_info

    def _process_one(self, tile_info):
        # --- Init
        processed_tile_filepath = os.path.join(self.processed_dirpath, tile_info["processed_relative_filepath"])
        processed_tile_dirpath = os.path.dirname(processed_tile_filepath)
        stats_filepath = os.path.join(processed_tile_dirpath, f"{tile_info['number']:08d}.stats.pt")
        os.makedirs(processed_tile_dirpath, exist_ok=True)

        # --- Check if tile has been processed already
        if os.path.exists(stats_filepath):
            stats = torch.load(stats_filepath)
            return stats

        tile_info = self.load_raw_data(tile_info)

        tile_info = self.pre_transform(tile_info)

        # Compute stats
        stats = {
            "mean": np.mean(tile_info["image"].reshape(-1, tile_info["image"].shape[-1]), axis=0) / 255,
            "var": np.var(tile_info["image"].reshape(-1, tile_info["image"].shape[-1]), axis=0) / 255,
            "class_freq": np.mean(tile_info["gt_polygons_image"], axis=(0, 1)) / 255,
            "disaster": tile_info["disaster"]  # Add disaster name to stats for aggregating per disaster
        }

        # Save data
        torch.save(tile_info, processed_tile_filepath)
        torch.save(stats, stats_filepath)

        return stats

    def __len__(self):
        if self.pre_process:
            return len(self.processed_relative_paths)
        else:
            return len(self.tile_info_list)

    def __getitem__(self, idx):
        if self.pre_process:
            filepath = os.path.join(self.processed_dirpath, self.processed_relative_paths[idx])
            data = torch.load(filepath)
            data["image_mean"] = self.stats["mean"][data["disaster"]]
            data["image_std"] = self.stats["std"][data["disaster"]]
            data["class_freq"] = self.stats["class_freq"]
        else:
            tile_info = self.tile_info_list[idx]
            # Load raw data
            data = self.load_raw_data(tile_info)
            raise NotImplementedError("Need to implement mean and std computation")

        # --- Crop to path_size
        height, width, _ = data["image"].shape
        pre_crop_image_norm = data["image"].shape[0] + data["image"].shape[1]
        crop_i = random.randint(0, height - self.patch_size)
        crop_j = random.randint(0, width - self.patch_size)
        data["image"] = data["image"][crop_i:crop_i + self.patch_size, crop_j:crop_j + self.patch_size]
        data["gt_polygons_image"] = data["gt_polygons_image"][crop_i:crop_i + self.patch_size, crop_j:crop_j + self.patch_size]
        data["gt_crossfield_angle"] = data["gt_crossfield_angle"][crop_i:crop_i + self.patch_size, crop_j:crop_j + self.patch_size]
        data["distances"] = data["distances"][crop_i:crop_i + self.patch_size, crop_j:crop_j + self.patch_size]
        data["sizes"] = data["sizes"][crop_i:crop_i + self.patch_size, crop_j:crop_j + self.patch_size]
        post_crop_image_norm = data["image"].shape[0] + data["image"].shape[1]
        # Sizes and distances are affected by cropping because they are relative to the image's norm (height + width).
        # All non-one pixels have to be renormalized:
        size_ratio = pre_crop_image_norm / post_crop_image_norm
        data["distances"][data["distances"] != 1] *= size_ratio
        data["sizes"][data["sizes"] != 1] *= size_ratio
        # ---

        data = self.transform(data)
        return data


def main():
    # Test using transforms from the frame_field_learning project:
    from frame_field_learning import data_transforms

    config = {
        "data_dir_candidates": [
            "/data/titane/user/nigirard/data",
            "~/data",
            "/data"
        ],
        "dataset_params": {
            "root_dirname": "xview2_xbd_dataset",
            "pre_process": True,
            "small": False,
            "data_patch_size": 725,
            "input_patch_size": 512,

            "train_fraction": 0.75
        },
        "num_workers": 8,
        "data_aug_params": {
            "enable": True,
            "vflip": True,
            "affine": True,
            "scaling": [0.9, 1.1],
            "color_jitter": True,
            "device": "cuda"
        }
    }

    # Find data_dir
    data_dir = python_utils.choose_first_existing_path(config["data_dir_candidates"])
    if data_dir is None:
        print_utils.print_error("ERROR: Data directory not found!")
        exit()
    else:
        print_utils.print_info("Using data from {}".format(data_dir))
    root_dir = os.path.join(data_dir, config["dataset_params"]["root_dirname"])

    # --- Transforms: --- #
    # --- pre-processing transform (done once then saved on disk):
    # --- Online transform done on the host (CPU):
    online_cpu_transform = data_transforms.get_online_cpu_transform(config,
                                                                    augmentations=config["data_aug_params"]["enable"])
    train_online_cuda_transform = data_transforms.get_online_cuda_transform(config,
                                                                            augmentations=config["data_aug_params"][
                                                                                "enable"])
    kwargs = {
        "pre_process": config["dataset_params"]["pre_process"],
        "transform": online_cpu_transform,
        "patch_size": config["dataset_params"]["data_patch_size"],
        "pre_transform": data_transforms.get_offline_transform_patch(),
        "small": config["dataset_params"]["small"],
        "pool_size": config["num_workers"],
    }
    # --- --- #
    fold = "train"
    if fold == "train":
        dataset = xView2Dataset(root_dir, fold="train", **kwargs)
    elif fold == "val":
        dataset = xView2Dataset(root_dir, fold="train", **kwargs)
    elif fold == "test":
        dataset = xView2Dataset(root_dir, fold="test", **kwargs)
    else:
        raise NotImplementedError

    print(f"dataset has {len(dataset)} samples.")
    print("# --- Sample 0 --- #")
    sample = dataset[0]
    for key, item in sample.items():
        print("{}: {}".format(key, type(item)))

    print("# --- Samples --- #")
    # for data in tqdm(dataset):
    #     pass

    data_loader = torch.utils.data.DataLoader(dataset, batch_size=1, shuffle=False, num_workers=config["num_workers"])
    print("# --- Batches --- #")
    for batch in tqdm(data_loader):

        # batch["distances"] = batch["distances"].float()
        # batch["sizes"] = batch["sizes"].float()

        # im = np.array(batch["image"][0])
        # im = np.moveaxis(im, 0, -1)
        # skimage.io.imsave('im_before_transform.png', im)
        #
        # distances = np.array(batch["distances"][0])
        # distances = np.moveaxis(distances, 0, -1)
        # skimage.io.imsave('distances_before_transform.png', distances)
        #
        # sizes = np.array(batch["sizes"][0])
        # sizes = np.moveaxis(sizes, 0, -1)
        # skimage.io.imsave('sizes_before_transform.png', sizes)

        print("----")
        print(batch["name"])

        print("image:", batch["image"].shape, batch["image"].min().item(), batch["image"].max().item())
        im = np.array(batch["image"][0])
        im = np.moveaxis(im, 0, -1)
        skimage.io.imsave('im.png', im)

        if "gt_polygons_image" in batch:
            print("gt_polygons_image:", batch["gt_polygons_image"].shape, batch["gt_polygons_image"].min().item(),
                  batch["gt_polygons_image"].max().item())
            seg = np.array(batch["gt_polygons_image"][0]) / 255
            seg = np.moveaxis(seg, 0, -1)
            seg_display = utils.get_seg_display(seg)
            seg_display = (seg_display * 255).astype(np.uint8)
            skimage.io.imsave("gt_seg.png", seg_display)

        if "gt_crossfield_angle" in batch:
            print("gt_crossfield_angle:", batch["gt_crossfield_angle"].shape, batch["gt_crossfield_angle"].min().item(),
                  batch["gt_crossfield_angle"].max().item())
            gt_crossfield_angle = np.array(batch["gt_crossfield_angle"][0])
            gt_crossfield_angle = np.moveaxis(gt_crossfield_angle, 0, -1)
            skimage.io.imsave('gt_crossfield_angle.png', gt_crossfield_angle)

        if "distances" in batch:
            print("distances:", batch["distances"].shape, batch["distances"].float().min().item(),
                  batch["distances"].float().max().item())
            distances = np.array(batch["distances"][0])
            distances = np.moveaxis(distances, 0, -1)
            skimage.io.imsave('distances.png', distances)

        if "sizes" in batch:
            print("sizes:", batch["sizes"].shape, batch["sizes"].float().min().item(), batch["sizes"].float().max().item())
            sizes = np.array(batch["sizes"][0])
            sizes = np.moveaxis(sizes, 0, -1)
            skimage.io.imsave('sizes.png', sizes)

        # valid_mask = np.array(batch["valid_mask"][0])
        # valid_mask = np.moveaxis(valid_mask, 0, -1)
        # skimage.io.imsave('valid_mask.png', valid_mask)

        input("Press enter to continue...")

        print("Apply online tranform:")
        batch = utils.batch_to_cuda(batch)
        batch = train_online_cuda_transform(batch)
        batch = utils.batch_to_cpu(batch)

        print("image:", batch["image"].shape, batch["image"].min().item(), batch["image"].max().item())
        print("gt_polygons_image:", batch["gt_polygons_image"].shape, batch["gt_polygons_image"].min().item(),
              batch["gt_polygons_image"].max().item())
        print("gt_crossfield_angle:", batch["gt_crossfield_angle"].shape, batch["gt_crossfield_angle"].min().item(),
              batch["gt_crossfield_angle"].max().item())
        # print("distances:", batch["distances"].shape, batch["distances"].min().item(), batch["distances"].max().item())
        # print("sizes:", batch["sizes"].shape, batch["sizes"].min().item(), batch["sizes"].max().item())

        # Save output to visualize
        seg = np.array(batch["gt_polygons_image"][0])
        seg = np.moveaxis(seg, 0, -1)
        seg_display = utils.get_seg_display(seg)
        seg_display = (seg_display * 255).astype(np.uint8)
        skimage.io.imsave("gt_seg.png", seg_display)

        im = np.array(batch["image"][0])
        im = np.moveaxis(im, 0, -1)
        skimage.io.imsave('im.png', im)

        gt_crossfield_angle = np.array(batch["gt_crossfield_angle"][0])
        gt_crossfield_angle = np.moveaxis(gt_crossfield_angle, 0, -1)
        skimage.io.imsave('gt_crossfield_angle.png', gt_crossfield_angle)

        distances = np.array(batch["distances"][0])
        distances = np.moveaxis(distances, 0, -1)
        skimage.io.imsave('distances.png', distances)

        sizes = np.array(batch["sizes"][0])
        sizes = np.moveaxis(sizes, 0, -1)
        skimage.io.imsave('sizes.png', sizes)

        # valid_mask = np.array(batch["valid_mask"][0])
        # valid_mask = np.moveaxis(valid_mask, 0, -1)
        # skimage.io.imsave('valid_mask.png', valid_mask)

        input("Press enter to continue...")


if __name__ == '__main__':
    main()



from lydorn_utils import polygon_utils


class RemoveDoubles(object):
    """Removes redundant vertices of all input polygons"""

    def __init__(self, epsilon=0.1):
        self.epsilon = epsilon

    def __call__(self, polygons):
        return [polygon_utils.remove_doubles(polygon, epsilon=self.epsilon) for polygon in polygons]
